# +
"""
Tools for finding and parsing placeholders described in the wiki of the library
"""
import json
import logging
from dataclasses import dataclass, field as dataclass_field
from typing import ClassVar, Tuple

# local imports
from npdoc_to_md.config import Config
from npdoc_to_md.logger import log


# -

# # Placeholder class

@dataclass
class Placeholder:
    """
    Helper for finding placeholders defined by this library in strings or markdown files.

    Attributes
    ----------
    required
        Mandatory parameters for placeholders
    optional
        Optional parameters for placeholders

    Examples
    --------
    >>> lines = ['Pandas is a cool library. Here is for instance the DataFrame object:',
    ...          '{{"obj":"pandas.DataFrame", "alias":"pd.DataFrame", "remove_doctest_blanklines":true}}']
    >>>
    >>> for line in lines:
    ...     placeholder = Placeholder.search(line)
    ...     if placeholder:
    ...         print(placeholder.obj_namespace, placeholder.config, sep=' | ')
    pandas.DataFrame | Config(alias='pd.DataFrame', examples_md_lang='python', remove_doctest_blanklines=True, \
remove_doctest_skip=False, md_section_level=3, ignore_custom_section_warning=False, members=[])
    """
    line:str
    parsed:dict = dataclass_field(init=False)
    obj_namespace:str = dataclass_field(init=False)
    config:Config = dataclass_field(init=False)
    required:ClassVar[Tuple[str]] = ('obj',)
    optional:ClassVar[Tuple[str]] = tuple(Config.__dataclass_fields__.keys())

    def __post_init__(self):
        line = self.line
        parsed:dict = json.loads(line[1:-1])

        # validate
        # 1) mandatory keys
        missing_required = [k for k in self.required if k not in parsed]
        if missing_required:
            raise ValueError(f'The mandatory keys {missing_required} are missing from a placeholder. '
                             f'The faulty line was:\n{line}')

        # 2) extra keys (not allowed)
        all_keys = list(self.required) + list(self.optional)
        extra_keys = [k for k in parsed if k not in all_keys]
        if extra_keys:
            raise ValueError(f'Unexpected keys {extra_keys} in a placeholder of numpydoc_to_md. '
                             f'Allowed keys are: {all_keys}. Faulty line:\n{line}')

        # create config object (which will validate it as well)
        obj_namespace = parsed['obj']  # don't pop "obj" we need it for attribute `parsed`
        kwargs = {k:v for k, v in parsed.items() if k != 'obj'}
        config = Config(**kwargs)

        # because we are in a frozen dataclass we need this workaround to set attributes
        object.__setattr__(self, 'parsed', parsed)
        object.__setattr__(self, 'obj_namespace', obj_namespace)
        object.__setattr__(self, 'config', config)

    @classmethod
    def search(cls, line:str) -> 'Placeholder':
        """
        If given line looks like a placeholder for our library (starts with "{{" and ends
        with "}}") we attempt to parse it (this may raise various exceptions) and return an instance
        of the class.
        Otherwise we return None.
        """
        line = line.strip()
        return cls(line=line) if line.startswith('{{') and line.endswith('}}') else None

    @classmethod
    def search_no_err(cls, line:str) -> 'Placeholder':
        """
        Same as method `search` (see its docstring) but when an exception occurs:
        * it is logged and not raised
        * None is returned

        Examples
        --------
        >>> line = '{{"obj":"datetime.datetime"}}'
        >>> placeholder = Placeholder.search_no_err(line=line)
        >>> print(placeholder.parsed)
        {'obj': 'datetime.datetime'}

        >>> # fails due to extra key
        >>> line = '{{"obj":"datetime.datetime", "invalid_key":"foo"}}'
        >>> placeholder = Placeholder.search_no_err(line=line)
        >>> print(placeholder)
        None

        >>> # fails due to missing mandatory "obj" key
        >>> line = '{{"not a placeholder":"foo"}}'
        >>> placeholder = Placeholder.search_no_err(line=line)
        >>> print(placeholder)
        None
        """
        line = line.strip()
        # case where it is definitely not a placeholder
        if not (line.startswith('{{') and line.endswith('}}')):
            return None
        # try to parse the placeholder
        try:
            return cls(line=line)
        except Exception:
            log(f'An exception occured when rendering this placeholder: {line}',
                level=logging.ERROR, exc_info=True)
            return None



# +
"""
Contains all objects that will be directly exposed to the users of the library
e.g. `render_obj_docstring`
"""
from functools import wraps
import logging
import re
import sys
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import List, Optional, Union

# local imports
from npdoc_to_md.helpers import FileOperations, Patterns
from npdoc_to_md.logger import log
from npdoc_to_md.parsers import parse_and_render
from npdoc_to_md.placeholder import Config, Placeholder


# -
# # Local helpers

# +
@dataclass(frozen=True)
class RenderedFile:
    """
    Represents a template file that has been converted by the npdoc_to_md library.

    The attributes listed below are also the parameters for making an instance.
    All parameters are mandatory.

    Note that for the attributes "_" and "-" are interchangeable (e.g. "original_text"
    and "original-text" will both work). This was made to simplify the CLI usage.

    Attributes
    ----------
    source
        Filepath to the template
    destination
        Filepath to the rendered template or None if the user did
        not wish to save the result to a file
    original_text
        Text of the template file
    rendered_text
        Text obtained after rendering the template file (after a function
        from npdoc_to_md was used)
    """
    source:str
    destination:Union[str, None]
    original_text:str
    rendered_text:str

    def __getattr__(self, attr:str):
        """
        We overwrite __getattr__ to allow aliases such as "original-text" for the CLI.
        """
        return self.__getattribute__(attr.replace('-', '_'))

    def __repr__(self):
        return (f'source: {self.source}\n'
                f'destination: {self.destination}\n'
                f'original_text: see attribute "original_text" (hidden to avoid cluttering the output)\n'
                f'rendered_text:\n{self.rendered_text}')


@dataclass(frozen=True)
class RenderedFileCLI(RenderedFile):
    """
    Same as class npdoc_to_md.RenderedFile (see its docstring)
    where the string form (str(instance)) has been changed for the CLI.
    Indeed, the string form will give the same result as the repr
    (repr(instance)).
    """

    def __str__(self):
        return super().__repr__()

    @classmethod
    def _from_rendered_file(cls, obj:RenderedFile):
        assert isinstance(obj, RenderedFile)
        return cls(**asdict(obj))


class RenderedFilesCLI(list):
    """
    Container for multiple instances of RenderedFileCLI.
    Behaves like a list but the string form (str(instance)) has been
    changed for the CLI.

    Examples
    --------
    >>> data = [RenderedFileCLI(source='test.md', destination='test.md', original_text='foo', rendered_text='foo')]
    >>> rfc = RenderedFilesCLI(data)
    >>> rfc
    [RenderedFileCLI(source='test.md', destination='test.md', original_text='foo', rendered_text='foo')]
    >>> len(rfc)
    1
    >>> rfc[0]
    RenderedFileCLI(source='test.md', destination='test.md', original_text='foo', rendered_text='foo')
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for obj in self:
            assert isinstance(obj, RenderedFileCLI)

    def __str__(self):  # pragma: no cover
        strings = []
        nb_files = len(self)
        for ix, f in enumerate(self):
            # make a separator so we can distinguish files more easely
            begin = '\n\n' if ix != 0 else ''
            strings.append(f'{begin}{"$"*50} FILE {ix + 1}/{nb_files} {"$"*50}\n')
            strings.append(f.__str__())
        return ''.join(strings)


# -

# # Render using Python objects

# IMPORTANT! further down the line (in other modules) `obj` will be referred to as `obj_namespace`
# and `obj` will be the actual Python object at given importable path
def render_obj_docstring(obj:str,
                         alias:Optional[str]=Config.get_default('alias'),
                         examples_md_lang:str=Config.get_default('examples_md_lang'),
                         remove_doctest_blanklines:bool=Config.get_default('remove_doctest_blanklines'),
                         remove_doctest_skip:bool=Config.get_default('remove_doctest_skip'),
                         md_section_level:int=Config.get_default('md_section_level'),
                         ignore_custom_section_warning:bool=Config.get_default('ignore_custom_section_warning'),
                         members:Optional[List[str]]=None) -> str:
    """
    Converts the docstring of an object (e.g. function, class, method, module, ...)
    to a pretty markdown string.

    It uses the same parameters as the ones for placeholders described
    in the wiki of the library.
    See wiki folder at the root of the repo or https://github.com/ThibTrip/npdoc_to_md/wiki

    CLI Examples
    ------------
    Note that "-" and "_" are interchangeable

    ```
    $ npdoc-to-md render-obj-docstring -obj "datetime.datetime" --alias "datetime" --examples-md-lang "raw" --md-section-level 3 --remove-doctest-skip --remove-doctest-blanklines

    $ npdoc-to-md render-obj-docstring -obj "pandas" --examples-md-lang "raw" --members "['DataFrame']"

    $ npdoc-to-md render-obj-docstring -obj "pandas.DataFrame" --alias "pd.DataFrame" --examples-md-lang "raw" --members "['public$', '+__dict__', '-to_dict']"
    ```

    Examples
    --------
    >>> from npdoc_to_md import render_obj_docstring
    >>>
    >>> md = render_obj_docstring(obj="npdoc_to_md.testing.now_utc",
    ...                           alias="now_utc",
    ...                           examples_md_lang='raw',
    ...                           remove_doctest_skip=True,
    ...                           remove_doctest_blanklines=True,
    ...                           md_section_level=3)
    """
    if not isinstance(obj, str):
        raise TypeError('Expected parameter `obj` to be a string of an importable python object '
                        f'e.g. "pandas.DataFrame", "datetime.datetime", ... . Got type {type(obj)} instead')

    config = Config(alias=alias,
                    examples_md_lang=examples_md_lang,
                    remove_doctest_blanklines=remove_doctest_blanklines,
                    remove_doctest_skip=remove_doctest_skip,
                    md_section_level=md_section_level,
                    ignore_custom_section_warning=ignore_custom_section_warning,
                    members=[] if members is None else members)
    return parse_and_render(obj_namespace=obj, config=config)


# # Render using text 

# ## Helpers to render a single placeholder

# +
def _render_placeholder(placeholder:Placeholder) -> str:
    r"""
    Uses instructions in given placeholder (syntax given by library)
    to find a Python object and render its docstring in Markdown.

    Examples
    --------
    >>> from npdoc_to_md.placeholder import Placeholder
    >>>
    >>> p = Placeholder('{{"obj":"npdoc_to_md.testing.now_utc", "alias":"now_utc", "examples_md_lang":"raw"}}')
    >>> md = _render_placeholder(placeholder=p)
    """
    config:Config = placeholder.config
    rendered = render_obj_docstring(obj=placeholder.obj_namespace,
                                    alias=config.alias,
                                    examples_md_lang=config.examples_md_lang,
                                    remove_doctest_blanklines=config.remove_doctest_blanklines,
                                    remove_doctest_skip=config.remove_doctest_skip,
                                    md_section_level=config.md_section_level,
                                    ignore_custom_section_warning=config.ignore_custom_section_warning,
                                    members=config.members)
    return rendered


def _render_placeholder_no_err(placeholder:Placeholder) -> str:
    """
    Similar to `_render_placeholder` but logs errors instead of raising
    them.

    Examples
    --------
    >>> from npdoc_to_md.placeholder import Placeholder
    >>>
    >>> # this is obviously not going to work (no module named monkey)
    >>> # but it will not raise any exception
    >>> p = Placeholder('{{"obj":"monkey"}}')
    >>> md = _render_placeholder_no_err(placeholder=p)
    >>>
    >>> # the placeholder is returned as is, no conversion was done
    >>> print(md)
    {{"obj":"monkey"}}
    """
    try:
        return _render_placeholder(placeholder=placeholder)
    except Exception:
        log(f'An exception occured when rendering this placeholder: {placeholder}',
            level=logging.ERROR, exc_info=True)
        return placeholder.line


# -

# ## Function to render from text

def render_string(string:str, ignore_errors:bool=False) -> str:
    r'''
    Returns a markdown string where placeholders defined in this library have
    been replaced with the corresponding docstrings rendered in Markdown.

    See wiki folder at the root of the repo or https://github.com/ThibTrip/npdoc_to_md/wiki
    and the examples for the placeholders syntax.

    Parameters
    ----------
    string
        String with markdown syntax (we will split it in lines)
        containing (or not) placeholders defined in this library
    ignore_errors
        If True only logs errors relative to converting placeholders
        in the string otherwise raises such errors.
        Also, placeholders that failed to convert will
        be unchanged in the resulting text

        Note that any string that begings with "{{" and ends with "}}" on the
        same line is interpreted as a placeholder by the library.

    CLI Examples
    ------------
    Note that "-" and "_" are interchangeable.

    While I do make this method available for the CLI, if you are going to do similar
    things as the examples below, I would suggest looking at the other functions of
    the library for generating documentation. Indeed, unless you are in Bash and can use a multiline string,
    the escaping is tricky as you can see below.

    IMPORTANT: do not put spaces in the placeholders (e.g. between commas) otherwise the parsing would fail (because
    space is a character used to separate arguments)

    Bash example:

    ```
    $ STRING='"{{\"obj\":\"npdoc_to_md.testing.DocumentedClassExample\",\"remove_doctest_blanklines\":true,\"remove_doctest_skip\":true,\"examples_md_lang\":\"raw\",\"members\":[\"public$\"]}}"'
    $ npdoc-to-md render-string -string "$STRING"
    ```

    Command Prompt example (Windows):

    ```
    $ set STRING='"{{\"obj\":\"npdoc_to_md.testing.DocumentedClassExample\",\"remove_doctest_blanklines\":true,\"remove_doctest_skip\":true,\"examples_md_lang\":\"raw\",\"members\":[\"public$\"]}}"'
    $ npdoc-to-md render-string -string %STRING%
    ```

    Bash example (multiline):

    ```
    $ STRING='Demo:
    > {{"obj":"npdoc_to_md.testing.DocumentedClassExample", "remove_doctest_blanklines":true, "md_section_level":3, "members":["public$"]}}'
    $ npdoc-to-md render-string -string "$STRING"
    ```

    Examples
    --------
    >>> from npdoc_to_md import render_string
    >>>
    >>> text = """
    ... Here is a demo of npdoc_to_md.render_string. The placeholder
    ... below will be converted to a docstring in markdown:
    ...
    ... ---
    ...
    ... {{"obj":"npdoc_to_md.testing.now_utc", "alias":"now_utc"}}
    ... """
    >>> md = render_string(string=text)

    >>> # demonstrating "ignore_errors": this raises no error even though we are referring to a non existent object
    >>> md = render_string(string='{{"obj":"some_object_that_does_not_exist"}}', ignore_errors=True)
    '''
    if ignore_errors:
        render_placeholder_method = _render_placeholder_no_err
        search_placeholder_method = Placeholder.search_no_err
    else:
        render_placeholder_method = _render_placeholder
        search_placeholder_method = Placeholder.search

    new_lines:List[str] = []
    for line in string.splitlines():
        placeholder:Union[Placeholder, None] = search_placeholder_method(line)
        if placeholder is not None:
            rendered = render_placeholder_method(placeholder=placeholder)
            new_lines.append(rendered)
        else:
            new_lines.append(line)
    return '\n'.join(new_lines)


# # Render using file

def render_file(source:str, destination:Optional[str]=None,
                ignore_errors:bool=False) -> Union[RenderedFile, RenderedFileCLI]:
    """
    Reads markdown file at path `source` and replaces placeholders defined in this library
    with the corresponding docstrings. It then returns the "converted" text.

    If a `destination` is given we save the converted text at this path.

    See wiki folder at the root of the repo or https://github.com/ThibTrip/npdoc_to_md/wiki
    for the placeholders syntax.

    Parameters
    ----------
    source
        Path to the markdown file
    destination
        If None only returns a converted markdown string.
        Otherwise it also creates or overwrites the file
        at this path with the converted markdown string.
    ignore_errors
        Same logic as in function `render_string` (see its docstring)

    CLI Examples
    ------------
    Note that "-" and "_" are interchangeable

    ```
    $ npdoc-to-md render-file -source "./docs/Render file.npmd" --destination "./docs/Render file.md"
    ```

    Since we are using python-fire for the CLI it is possible to access attributes
    of the class (RenderedFile) instance that is returned. In the example below
    we are acccessing the attribute `rendered_text`:

    ```
    $ npdoc-to-md render-file -source "./docs/Render file.npmd" --destination "./docs/Render file.md" - rendered-text
    ```

    See also the python-fire guide: https://github.com/google/python-fire/blob/master/docs/guide.md

    Returns
    -------
    npdoc_to_md.RenderedFile | npdoc_to_md.RenderedFileCLI
        Returns RenderedFile object if using this function in Python.
        Returns RenderedFileCLI if using this function in CLI.
        See docstring of these objects.

    Examples
    --------
    >>> from npdoc_to_md import render_file
    >>>
    >>> source = "./README.md"
    >>> destination = "./README_converted.md"
    >>> rendered_file = render_file(source=source, destination=destination) # doctest: +SKIP
    """
    # read contents
    log(f'Reading file contents at path {source}')
    with open(source, mode='r', encoding='utf-8', newline='\n') as fh:
        original_text = fh.read()
    # render
    log('Rendering file contents')
    rendered_text = render_string(string=original_text, ignore_errors=ignore_errors)

    # save
    if destination is not None:
        log(f'Saving rendered contents to path {destination}')
        with open(destination, mode='w', encoding='utf-8', newline='\n') as fh:
            fh.write(rendered_text)
    return RenderedFile(source=source, destination=destination,
                        original_text=original_text, rendered_text=rendered_text)


# # Render using folder

def render_folder(source:str,
                  destination:Optional[str]=None,
                  recursive:bool=False,
                  ignore_errors:bool=False,
                  pattern:Optional[str]=None,
                  case_sensitive:bool=False) -> Union[List[RenderedFile], RenderedFilesCLI]:
    r"""
    Reads all markdown files in the folder at path `source`
    and for each markdown file, replaces placeholders defined in this library
    with the corresponding docstrings.

    If a `destination` folder is given we also save converted files there using
    the same folder structure (case of `recursive=True`) and also the same
    file names, except the extensions that always get converted to ".md"
    (unless they already matched ".md" case insensitive) e.g.:
    * `some_lib/docs_templates/Home.npmd` -> `some_lib/docs/Home.md`
    * `some_lib/docs_templates/Logging.MD` -> `some_lib/docs/Logging.MD`
    * `some_lib/docs_templates/cool_subpackage/Home.npmd` -> `some_lib/docs/cool_subpackage/Home.md`

    See wiki folder at the root of the repo or https://github.com/ThibTrip/npdoc_to_md/wiki
    for the placeholders syntax.

    Parameters
    ----------
    source
        Path to the folder containing markdown files
    destination
        If None no file operations is done but you still get the result
        of the rendering (see sections Examples | CLI Examples and Returns)
    recursive
        If False only looks for markdown files directly in folder `source`
        otherwise also looks inside subfolders
    ignore_errors
        Same logic as in function `render_string` (see its docstring)
    pattern
        Regex pattern for matching file names in folder `source`
        e.g. "(\.md|\.txt)$" (file names with the extension ".md"
        or ".txt")

        By default we use a regex pattern matching "md" and "npmd"
        extensions. See attributes `template_files` and `template_files_insensitive`
        of class `npdoc_to_md.helpers.Patterns`
    case_sensitive
        Whether the `pattern` is case sensitive. By default this is False (`pattern` is
        case insensitive)

    CLI Examples
    ------------
    Note that "-" and "_" are interchangeable

    ```
    $ npdoc-to-md render-folder -source "./docs" --destination "./docs" --ignore-errors --recursive
    ```

    Since we are using python-fire for the CLI it is possible to work with the returned object (a list
    of RenderedFileCLI instances). In the example below we are acccessing the attribute `rendered_text`
    of the first item:

    ```
    $ npdoc-to-md render-folder -source "./docs" --destination "./docs" --ignore-errors --recursive - 0 - rendered_text
    ```

    See also the python-fire guide: https://github.com/google/python-fire/blob/master/docs/guide.md

    Returns
    -------
    list[npdoc_to_md.RenderedFile] | RenderedFilesCLI
        Returns list[RenderedFile] if using this function in Python.
        Returns RenderedFilesCLI if using this function in the CLI.

        See docstring of these objects.

    Examples
    --------
    >>> from npdoc_to_md import render_folder
    >>>
    >>> source = "./docs_templates"
    >>> destination = "./docs"
    >>> rendered_files = render_folder(source=source, destination=destination) # doctest: +SKIP
    """
    # get pattern object for extensions
    if pattern is None:
        pattern:re.Pattern = Patterns.template_files if case_sensitive else Patterns.template_files_insensitive
    else:
        flags = re.UNICODE if case_sensitive else re.IGNORECASE | re.UNICODE
        pattern = re.compile(pattern, flags=flags)

    # get markdown files in folder
    filepaths = FileOperations.list_files(folder=source, pattern=pattern, recursive=recursive)
    nb_files = len(filepaths)

    # render markdown files
    rendered_files = []
    log(f'{"#"*10} Processing folder "{source}" {"#"*10}')
    for ix, filepath in enumerate(filepaths):
        log(f'{"-"*10} Processing path {ix+1}/{nb_files} {"-"*10}')
        if destination is not None:
            destination_path = FileOperations.switch_folder(source_folder=source,
                                                            destination_folder=destination,
                                                            filepath=filepath)
            # replace extension to Markdown (case of template or otherwise)
            not_md = not destination_path.lower().strip().endswith('.md')
            if not_md:
                destination_path = str(Path(destination_path).with_suffix('.md'))
        else:
            destination_path = None
        rendered_file = render_file(source=filepath, destination=destination_path, ignore_errors=ignore_errors)
        rendered_files.append(rendered_file)
    return rendered_files


# # Start CLI

def start_cli() -> None:
    """
    Starts the Command Line Interface for the library.
    Is also the entry point in the file `setup.py`
    """
    try:
        import fire
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError('Please install optional dependency "fire" (pip install fire) '
                                  'if you want to use the CLI') from e

    # change return types for render_file and render_folder
    def rendered_files_to_rendered_files_cli(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, RenderedFile):
                return RenderedFileCLI._from_rendered_file(obj=result)
            elif isinstance(result, list):
                assert all(isinstance(f, RenderedFile) for f in result)
                return RenderedFilesCLI(RenderedFileCLI._from_rendered_file(obj=f) for f in result)
            else:
                raise TypeError(f'Unexpected return type from function {func.__name__}: {type(result)}')
        return wrapper

    # cli mapping (which subcommand executes which function)
    cli_mapping = {'render-obj-docstring':render_obj_docstring,
                   'render-string':render_string,
                   'render-file':rendered_files_to_rendered_files_cli(render_file),
                   'render-folder':rendered_files_to_rendered_files_cli(render_folder)}

    # make it possible to use hyphens or underscore for subcommand
    # (I prefer transforming the arg to adding subcommands in the `cli_mapping`
    # because this would clutter the help)
    def hyphenize_subcommand():
        # check if argument was passed
        subcommand_ix = 1
        arg_passed = len(sys.argv) >= subcommand_ix + 1
        if not arg_passed:  # pragma: no cover
            return
        # check if argument has an underscore
        arg = sys.argv[subcommand_ix]
        if '_' not in arg:
            return
        # change the argument and check if it corresponds to any subcommand
        aliased = arg.replace('_', '-')
        if aliased in cli_mapping:
            log(f'Transformed subcommand: {arg} -> {aliased}', level=logging.DEBUG)
            sys.argv[subcommand_ix] = aliased

    hyphenize_subcommand()

    # fire CLI
    fire.Fire(cli_mapping)



"""
Contains configuration objects for functions of this library
as well as placeholders in template files
"""
from dataclasses import dataclass, field as dataclass_field, fields as dataclass_fields, MISSING
from enum import Enum
from typing import get_type_hints, List, Optional

# # Helpers

default_language_examples = 'python'


# ## Enum for special values to designate class members

class MemberFlag(Enum):
    """
    Special values to designate members of a Python object (see parameter `members`
    in the library documentation in the section for placeholders)
    """
    DUNDER = 'dunder$'
    PRIVATE = 'private$'
    PUBLIC = 'public$'


# ## Enum for special values to render examples

class SpecialExampleOutputLanguages(Enum):
    """
    Special ways of representing example outputs in
    a Python's docstring

    Attributes
    ----------
    DEFAULT : str
        Default language for syntax highlighting
        of example output within markdown code block
    RAW : str
        Flag for when we'll use a raw markdown code block
        (no language defined for syntax highlighting)
        for the output of an example
    MARKDOWN_RENDERED : str
        Flag for when we will **not** use a markdown code block
        for the output of an example (e.g. if the user
        wants to have a markdown table as output)
    """
    DEFAULT = default_language_examples
    RAW = 'raw'
    MARKDOWN_RENDERED = 'markdown_rendered'


# # Config class

@dataclass(frozen=True)
class Config:
    """
    Options found throughout the library for how to
    display docstrings in Markdown (see parameters
    for placeholders in the library documentation)
    """
    alias:Optional[str] = None
    examples_md_lang:str = default_language_examples
    remove_doctest_blanklines:bool = False
    remove_doctest_skip:bool = False
    md_section_level:int = 3
    ignore_custom_section_warning:bool = False
    members:List[str] = dataclass_field(default_factory=list)

    def __post_init__(self):
        # verify types
        config_types = get_type_hints(Config)
        for field in dataclass_fields(self):
            attr = field.name
            val = getattr(self, attr)
            type_ = type(val)

            # handle special cases (non primitives e.g. List[str])
            if attr == 'alias':
                if not isinstance(val, str) and val is not None:
                    raise TypeError(f'Parameter "{attr}" is not a str or None. Type: {type_}')
            elif attr == 'members':
                if not isinstance(val, list):
                    raise TypeError(f'Parameter "{attr}" is not a list. Type: {type_}')
                elif any(not isinstance(v, str) for v in val):
                    raise TypeError(f'The parameter "{attr}" (list) contains non strings.')
            # handle normal case
            else:
                expected_type = config_types[attr]
                if not isinstance(val, expected_type):
                    raise TypeError(f'Parameter "{attr}" is not of type {expected_type}. Type: {type_}')

    @staticmethod
    def get_default(field:str):
        """
        Gets the default value of given field of the dataclass.

        Examples
        --------
        >>> Config.get_default('examples_md_lang')
        'python'
        """
        try:
            default = Config.__dataclass_fields__[field].default
            if default is MISSING:  # pragma: no cover
                raise ValueError(f'Field "{field}" has no default value')
            return default
        except KeyError as e:
            raise AttributeError(f'Config has no attribute "{field}"') from e



# +
"""
Custom exceptions for the library
"""


class NonExistentObjectException(Exception):
    pass


class SignatureNotFoundException(Exception):
    pass


class NonExistentMemberException(Exception):
    pass


class MembersConflictsException(Exception):
    pass


class InvalidMembersFlagException(Exception):
    pass



"""
Dummy objects (classes, functions, ...) for demos of the library and tests

Notes
-----
The docstring of this module will be used for docstring rendering tests in pytest.
Which is why I am also going to include some examples here and a custom section
"Functions".

Examples
--------
>>> from npdoc_to_md.testing import DocumentedClassExample
>>>
>>> ex = DocumentedClassExample(first_name='John', last_name='Doe')
>>> ex.full_name
'John Doe'

Functions
---------
* create_random_table
  Returns a random table (pandas DataFrame)
* now_utc
  Returns the current date and time in UTC
* ...
"""
import datetime
from dataclasses import dataclass, fields as dataclassfields
from numpydoc.docscrape import NumpyDocString
from typing import Dict, List, Optional, Union


# # Dummy objects

# +
def create_random_table(nb_rows:int=5, seed:Optional[int]=None) -> 'pandas.DataFrame':
    """
    Returns a random table (pandas DataFrame). You must
    have pandas (and numpy which comes with it) installed!

    Parameters
    ----------
    nb_rows
        How many rows to generate
    seed
        Any number for having predictable outputs

    Examples
    --------
    >>> df = create_random_table(seed=0)
    >>> print(df.to_markdown(index=False))
    |   A |   B |   C |   D |
    |----:|----:|----:|----:|
    |  44 |  47 |  64 |  67 |
    |  67 |   9 |  83 |  21 |
    |  36 |  87 |  70 |  88 |
    |  88 |  12 |  58 |  65 |
    |  39 |  87 |  46 |  88 |
    """
    try:
        import pandas as pd
        import numpy as np
    except (ModuleNotFoundError, ImportError) as e:  # pragma: no cover
        raise ModuleNotFoundError('You need to install the optional dependency pandas for this '
                                  'test function (pip install pandas)') from e

    if seed is not None:
        np.random.seed(seed)
    return pd.DataFrame(np.random.randint(0, 100, size=(nb_rows, 4)),
                        columns=list('ABCD'))


def now_utc() -> datetime.datetime:
    """
    Returns the current date and time in UTC

    Examples
    --------
    >>> current_dt_utc = now_utc()
    >>> current_dt_utc  # doctest: +SKIP
    datetime.datetime(2022, 4, 14, 19, 46, 6, 504371, tzinfo=datetime.timezone.utc)
    """
    return datetime.datetime.now().astimezone(datetime.timezone.utc)


# -

# ## Functions and classes with empty docstrings/signatures/sections
#
# To try to break the library

# +
def empty_function():
    pass  # pragma: no cover


class EmptyClass:
    pass  # pragma: no cover


# make a function and a class with empty sections in the docstrings
# note that we necessarily need a line break between sections
# otherwise numpydoc will think everything belongs to the
# first section
def make_docstring_with_empty_sections():
    invisible_sections = ('Signature', 'Summary', 'Extended Summary')
    # since we cannot have both Yields and Returns we will remove Returns
    # also "index" is not a section but a method listed under NumpyDocString.sections
    irrelevant_sections = ('index', 'Returns')
    lines = []
    for section in NumpyDocString.sections:
        if section in invisible_sections + irrelevant_sections:
            continue
        lines.extend([section, '-' * len(section), ''])
    return '\n'.join(lines)


def empty_doc_sections_function():
    pass  # pragma: no cover


class EmptyDocSectionsClass:
    pass  # pragma: no cover


empty_doc_sections_function.__doc__ = make_docstring_with_empty_sections()
EmptyDocSectionsClass.__doc__ = make_docstring_with_empty_sections()


# -

# # Example of fully documented objects

# ## Functions

# +
def documented_func_example(a:int, b:int, *,
                            # add a bunch of stupid kwargs to try to break the library
                            # especially the escaping of special characters
                            useless_param=None,
                            _bad_name_param1=None,
                            __bad_name_param2=None,
                            # note: do not try Optional[str] as this gets converted to Union[str, None]
                            # in Python 3.7
                            crazy_typing_param:Union[Dict[str, int], List[str]]=None,
                            **kwargs) -> int:
    r"""
    Example function that sums to integers. This function is used for pytest.
    It does not really make sense, this is just for testing docstring rendering.

    Some extended summary.

    Multiline test.

    Custom Section
    --------------
    This is for testing that custom sections are not ignored (as opposed to what
    numpydoc does) and we can fetch the content in the right order.

    Multiline test.

    Notes
    -----
    Some notes.

    Multiline test.

    Warnings
    --------
    Cautionnary notes for the user.
    More warning notes.

    Multiline test.

    References
    ----------
    [0] Me

    Multiline test.

    See Also
    --------
    empty_function, EmptyClass, empty_doc_sections_function, EmptyDocSectionsClass
        Other useless functions.

        Numpydoc deletes line breaks for section "See Also" so no multiline test here...
        The next function I put in this section will have no description.
    now_utc

    Parameters
    ----------
    a : int
        First number.

        Multiline test.
    b : int

    Other Parameters
    ----------------
    useless_param
        Some useless parameter.

        Multiline test.

    _bad_name_param1
    __bad_name_param2
    crazy_typing_param
        Parameter with complex typing
    optional_type_param
        Parameter with optional typing

    Examples
    --------
    We are going to try to break our library by:
    1. Putting text before examples or not
    2. Separating examples by line breaks (also with irrelevant leading spaces) or not

    We'll then check if the library still manages to separate examples properly

    * example 1
    {{python}}
    >>> documented_func_example(1, 1)
    2

    * example 2
    {{python}}
    >>> documented_func_example(2, 2)
    4

    ---

    #### Examples unrelated to the function

    * test rendered markdown
    {{markdown_rendered}}
    >>> import pandas as pd
    >>> df = pd.DataFrame({'A':[1, 2]})
    >>> print(df.to_markdown())
    |    |   A |
    |---:|----:|
    |  0 |   1 |
    |  1 |   2 |
    
    * test raw code block
    {{raw}}
    >>> print('Hello world!')
    Hello world!

    * test non rendered markdown
    {{markdown}}
    >>> print('# some markdown string\n* test')
    # some markdown string
    * test

    >>> print('test')
    test
    >>> print('foo\n')
    foo
    <BLANKLINE>

    Returns
    -------
    c : int
        Result of a + b.

        Multiline test.

    Raises
    ------
    TypeError
        If a or b is not a number.

        Multiline test.

    Warns
    -----
    UserWarning
        Never occurs (this is just an example)

        Multiline test.
    """
    return a + b


def documented_generator_func_example(nb_vals=10):
    """
    Generator example (for testing the parsing of Yields and Receives sections).
    Simply yields a range from 0 to `nb_vals`.

    Receives
    --------
    nb_vals : int, default 10
        Number of values to yield.

        Multiline test.

    Yields
    ------
    i : int
        Incrementing integers.

        Multiline test.

    Examples
    --------
    >>> for i in documented_generator_func_example(2):
    ...    print(i)
    0
    1
    """
    for i in range(nb_vals):
        yield i


# -

# ## Classes

@dataclass(frozen=True)
class DocumentedClassExample:
    """
    Example dataclass representing a person. This class is used for pytest.
    It does not really make sense, this is just for testing docstring rendering.

    Attributes
    ----------
    full_name : str
        first name + last name (both names in title case and separated by a space).

        Multiline test.

    Examples
    --------
    >>> ex = DocumentedClassExample(first_name='John', last_name='Doe')
    >>> ex.full_name
    'John Doe'
    """
    first_name:str
    last_name:str

    @property
    def full_name(self) -> str:
        """
        Returns the full name using the names part defined in the instance.
        This is a test for rendering docstrings of properties.

        Returns
        -------
        str
            Full name

        Examples
        --------
        >>> ex = DocumentedClassExample(first_name='john', last_name='doe')
        >>> ex.full_name
        'John Doe'
        """
        return f'{self.first_name.title()} {self.last_name.title()}'

    def example_method(self) -> None:
        """
        Prints the first name defined in the instance.
        This is a test for rendering docstrings of methods.

        Examples
        --------
        >>> d = DocumentedClassExample('John', 'Doe')
        >>> d.example_method()
        John
        """
        print(self.first_name)

    @staticmethod
    def example_static_method() -> List[str]:
        """
        Prints the names of the dataclass fields.
        This is a test for rendering docstrings of staticmethods.

        Examples
        --------
        >>> DocumentedClassExample.example_static_method()
        ['first_name', 'last_name']

        >>> d = DocumentedClassExample('John', 'Doe')
        >>> d.example_static_method()
        ['first_name', 'last_name']
        """
        fields = dataclassfields(DocumentedClassExample)
        return [f.name for f in fields]

    @classmethod
    def test_class_method(cls) -> 'DocumentedClassExample':
        """
        Returns a "John Doe" instance of the class.
        This is a test for rendering docstrings of class methods.

        Examples
        --------
        >>> DocumentedClassExample.test_class_method()
        DocumentedClassExample(first_name='John', last_name='Doe')
        """
        return cls(first_name='John', last_name='Doe')

    def _example_private_method(self) -> None:
        """
        Some private method
        """
        pass  # pragma: no cover



# +
"""
Tests the helpers module
"""
import pytest
import re
import os
import tempfile
from pathlib import Path

# local imports
from npdoc_to_md.helpers import FileOperations, Patterns


# -

# # Test patterns

@pytest.mark.parametrize('pattern_attr, string, expected', [('console_py', '>>> print("hello")', '>>> '),
                                                            ('console_py', '...     2 + 2', '... '),
                                                            ('doctest_skip', '>>> print("hello") #doctest: +SKIP', ' #doctest: +SKIP'),
                                                            ('blankline', ' <BLANKLINE> ', ' <BLANKLINE> '),
                                                            ('self_or_cls', 'Foo(self, a:int, b:int)', 'self, '),
                                                            ('template_files', 'README.md', '.md'),
                                                            ('template_files', 'README.MD', None),  # does not match
                                                            ('template_files_insensitive', 'README.md', '.md'),
                                                            ('template_files_insensitive', 'README.MD', '.MD')])
def test_patterns(pattern_attr, string, expected):
    """
    Tests regex patterns listed as attributes in class npdoc_to_md.helpers.Patterns
    """
    pat:re.Pattern = getattr(Patterns, pattern_attr)
    if expected is None:
        assert not pat.search(string)
    else:
        assert pat.search(string).group(0) == expected


# # Test file listing

# +
test_pattern_file_listing = re.compile(r'\.txt$')


@pytest.mark.parametrize('recursive', [True, False], ids=['recursive', 'non recursive'])
def test_file_listing(recursive):
    with tempfile.TemporaryDirectory() as tmpdirpath:

        # create test structure
        # we will have a file 1 under root and a file 2 under a subfolder called "subfolder"
        filename1, subfoldername, filename2 = 'foo.txt', 'subfolder', 'bar.txt'
        with open(os.path.join(tmpdirpath, filename1), mode='w', encoding='utf-8') as fh:
            fh.write('test')
        os.mkdir(os.path.join(tmpdirpath, subfoldername))
        with open(os.path.join(tmpdirpath, subfoldername, filename2), mode='w', encoding='utf-8') as fh:
            fh.write('test')

        # list paths
        filepaths = FileOperations.list_files(folder=tmpdirpath, recursive=recursive, pattern=test_pattern_file_listing)
        nb_items_expected = 2 if recursive else 1
        assert len(filepaths) == nb_items_expected

        # check paths integrity

        # check file 1 is present
        assert any(Path(f).name == filename1 for f in filepaths)

        # check file 2 is present (when recursive) and check its parents
        if not recursive:
            return
        candidates = list(filter(lambda f: Path(f).name == filename2, filepaths))
        if len(candidates) == 0:
            raise AssertionError(f'Filepath of second test file seems absent. All paths:\n{filepaths}')  # pragma: no cover
        assert len(candidates) == 1, 'Multiple paths for second test file ?!'
        path_obj = Path(candidates[0])
        assert path_obj.parent.name == subfoldername
        assert path_obj.parent.parent.name == Path(tmpdirpath).name


# -

# # Test switching folders

# +
params_switch_folders = [
    ('D:/', 'C:/Test', 'D:/foo.md', 'C:/Test/foo.md'),
    ('D:/', 'C:/Test', 'D:/Subfolder/foo.md', 'C:/Test/Subfolder/foo.md'),
    ('/home/my_lib/docs', '/home/test', '/home/my_lib/docs/foo.md', '/home/test/foo.md'),
    ('/home/my_lib/docs', '/home/test', '/home/my_lib/docs/subfolder/foo.md', '/home/test/subfolder/foo.md')
]


@pytest.mark.parametrize('source, destination, filepath, expected', params_switch_folders)
def test_switch_folder(source, destination, filepath, expected):
    new_filepath = FileOperations.switch_folder(source_folder=source, destination_folder=destination,
                                                filepath=filepath, create_missing_dirs=False)
    # Path allows us to make OS agnostic comparisons
    assert Path(new_filepath) == Path(expected)



"""
Custom logger for the library

Module heavily inspired from:
https://github.com/SergeyPirogov/webdriver_manager/blob/master/webdriver_manager/logger.py

Also thanks to this post for coloring the logs:
https://stackoverflow.com/a/56944256/10551772
"""
import logging
import os
from dataclasses import dataclass


# # Helpers

# +
@dataclass(frozen=True)
class LoggingLevel:
    method_name:str
    color:str


# string for resetting color that we will need to use for every log line
# (in the formatting as we do not want to change the message)
reset_color = "\x1b[0m"


# mapping/switch {level:LoggingLevel}
loggers = {}
log_method_switch = {logging.CRITICAL:LoggingLevel(method_name='critical', color="\x1b[31;1m"),  # bold red
                     logging.ERROR:LoggingLevel(method_name='error', color="\x1b[31;20m"),  # red
                     logging.WARNING:LoggingLevel(method_name='warning', color="\x1b[33;20m"),  # yellow
                     logging.INFO:LoggingLevel(method_name='info', color="\x1b[38;20m"),  # grey
                     logging.DEBUG:LoggingLevel(method_name='debug', color="\x1b[1;34m")}  # blue


# -

# # Main function

def log(text:str, name:str='npdoc_to_md', level:int=logging.INFO, exc_info:bool=False) -> None:
    """
    Logs given text to stderr (default of the logging library).

    Parameters
    ----------
    text
        Text to log
    name
        Name of the logger
    level
        See logging levels
        https://docs.python.org/3/library/logging.html#logging-levels
    exc_info
        Whether to include exceptions info

    Notes
    -----
    This is heavily inspired from:
    https://github.com/SergeyPirogov/webdriver_manager/blob/master/webdriver_manager/logger.py

    Examples
    --------
    * setting the log level of the library via an environment variable
    >>> import os, logging
    >>> from npdoc_to_md.logger import log
    >>> os.environ['NPDOC_TO_MD_LOG_LEVEL'] = str(logging.WARNING) # doctest: +SKIP
    >>>
    >>> # this won't log anything (INFO level < WARNING level)
    >>> log('info', level=logging.INFO) # doctest: +SKIP
    >>>
    >>> # this will log something
    >>> log('warn', level=logging.WARNING) # doctest: +SKIP
    """
    # get the appropriate log method (info, warning etc.)
    try:
        log_level:LoggingLevel = log_method_switch[level]
    except KeyError:  # pragma: no cover
        raise ValueError(f'{level} is not a valid log level. See https://docs.python.org/3/library/logging.html')

    # environment variable so user can customize the logging level of the library
    logger_level = os.getenv('NPDOC_TO_MD_LOG_LEVEL', logging.INFO)
    logger_level = int(logger_level) if isinstance(logger_level, str) else logger_level

    # init logger
    if name not in loggers:
        _logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(color)s%(asctime)s | %(levelname)s     '
                                      '| %(name)s    | %(module)s:%(funcName)s:%(lineno)s '
                                      f'- %(message)s{reset_color}')
        handler.setFormatter(formatter)
        _logger.addHandler(handler)
        _logger.setLevel(logger_level)
        loggers[name] = _logger

    # log
    getattr(loggers[name], log_level.method_name)(msg=text, exc_info=exc_info, extra={'color': log_level.color})



# +
"""
Contains parsers for every sections listed by numpydoc and also a main
function `parse_and_render` that uses these parsers to convert docstrings
to markdown.

See sections in https://numpydoc.readthedocs.io/en/latest/format.html
"""
import dataclasses
import inspect
import pydoc
from abc import ABC, abstractmethod
from dataclasses import dataclass, field as dataclassfield
from numpydoc.docscrape import Parameter
from typing import Any, ClassVar, List, Tuple

# local imports
from npdoc_to_md.exceptions import (MembersConflictsException,
                                    InvalidMembersFlagException,
                                    NonExistentMemberException,
                                    NonExistentObjectException,
                                    SignatureNotFoundException)
from npdoc_to_md.helpers import Patterns, unique
from npdoc_to_md.examples_labeller import (ExampleBlock, ExamplesLabeller,
                                           ExampleLine, ExampleLineType)
from npdoc_to_md.logger import log
from npdoc_to_md.config import MemberFlag, Config, SpecialExampleOutputLanguages
from npdoc_to_md.sections import SectionsFinder


# -

# # Helpers

# ## Generic helpers

# +
class MdEscaper:
    r"""
    Escapes problematic characters for strings that are supposed
    to be rendered "as-is" and not in Markdown e.g. a function
    signature.

    Examples
    --------
    >>> MdEscaper.escape('foo(self, *some_args)')
    'foo(self, \\*some\\_args)'
    """
    translator = {ord(c):f'\\{c}' for c in ('_', '*')}

    def escape(s:str) -> str:
        return s.translate(MdEscaper.translator)


def param_description_lines_to_md_lines(desc_lines:List[str]) -> List[str]:
    """
    Formats the descriptions found below parameter names (numpydoc sections "Parameters",
    "Other Parameters", "Returns", "Yield", ...) or function names (numpydoc sections
    "See Also", ...).

    Notes
    -----
    After some trials I decided to format these lines by only adding 2 leading spaces because:
    * 1 leading space: it is ignored
    * 2 to 5: makes a bullet level indent (for all values so might as well use 2 spaces)
    * 6+ spaces: always makes a code block which breaks markdown formatting such as bullets :(

    Note that using &nbsp; and such is bad idea because we'd have to understand when markdown
    needs a new line or not (with tables and such it gets complicated quickly...)
    """
    return [(' ' * 2) + line if line.strip() else line for line in desc_lines]


# -

# ## Parser for the members of a Python object

@dataclass(frozen=True)
class MembersParser:
    """
    Parses the values in the field "members" of the placeholders
    as described in the wiki.
    See also Examples which are quite self-explanatory.

    Parameters
    ----------
    obj
        A Python object
    members
        Raw values of the field "members" in a placeholder

    Attributes
    ----------
    parsed_members
        Result obtained after parsing `members`: all the
        members that we will have to fetch the docstring
        from

    Examples
    --------
    >>> from npdoc_to_md.config import MemberFlag
    >>>
    >>> class Foo():
    ...     def __init__(self): pass
    ...     def _private_method1(self): pass
    ...     def _private_method2(self): pass
    ...     def __dunder_method1__(self): pass
    ...     def __dunder_method2__(self): pass
    ...     def public_method1(self): pass
    ...     def public_method2(self): pass
    >>>
    >>> public = MemberFlag.PUBLIC.value
    >>> private = MemberFlag.PRIVATE.value
    >>> dunder = MemberFlag.DUNDER.value
    >>>
    >>> MembersParser(obj=Foo, members=[public, '-public_method1', private]).parsed_members
    ('public_method2', '_private_method1', '_private_method2')

    >>> MembersParser(obj=Foo, members=['+public_method1', 'public_method2', dunder]).parsed_members[0:5]
    ('public_method1', 'public_method2', '__class__', '__delattr__', '__dict__')
    """
    obj:Any
    members:List[str]
    members_parsed:Tuple[str] = dataclassfield(init=False)
    allowed_flags:ClassVar[Tuple[str]] = tuple(f.value for f in MemberFlag)
    _all_members:List[str] = dataclassfield(init=False, repr=False)

    def _parse_flag(self, flag:str) -> List[str]:
        """
        Helper for __post_init__
        """
        if flag == MemberFlag.DUNDER.value:
            return [v for v in self._all_members if v.startswith('__')]
        elif flag == MemberFlag.PRIVATE.value:
            return [v for v in self._all_members if v.startswith('_') and not v.startswith('__')]
        elif flag == MemberFlag.PUBLIC.value:
            return [v for v in self._all_members if not v.startswith('_')]
        else:
            raise InvalidMembersFlagException('Flags for documenting members must be one of: '
                                              f'{self.allowed_flags}. Got: "{flag}"')

    def _parse_member_to_include(self, value:str) -> str:
        """
        Helper for __post_init__
        """
        # handle case of included with "+" syntax
        value = value[1:] if value.startswith('+') else value
        if value not in self._all_members:
            raise NonExistentMemberException(f'Object {self.obj} does not have any attribute: "{value}"')
        return value

    def __post_init__(self):
        # clean members listed by the users
        members = [m.strip() for m in self.members]
        if any(m == '' for m in members):
            raise ValueError('An empty string is not a valid member name for a Python object. '
                             f'Value of parameter `members` was: {members}')
        object.__setattr__(self, 'members', members)
        object.__setattr__(self, '_all_members', dir(self.obj))  # this is for avoiding many calls to `dir`

        candidates = []
        exclusions = []
        inclusions = []
        for m in members:
            if m in self.allowed_flags:
                candidates.extend(self._parse_flag(m))
            elif m.startswith('-'):
                exclusions.append(m[1:])
            else:
                # case "+member" or just "member"
                parsed = self._parse_member_to_include(m)
                candidates.append(parsed)
                inclusions.append(parsed)

        # case when we have members both excluded and included
        conflicts = set(exclusions) & set(inclusions)
        if conflicts:
            raise MembersConflictsException(f'The following members in object {self.obj} are set to be '
                                            f'included AND excluded: {conflicts}')

        # finalize parsing
        parsed_members = [c for c in unique(candidates) if c not in exclusions]
        object.__setattr__(self, 'parsed_members', tuple(parsed_members))


# ## Base parsers for numpy sections

# +
@dataclass(frozen=True)
class DocParser(ABC):
    '''
    Abstract class for taking information from a numpydoc section (numpydoc
    always give back a list of items for each section) and converting it to
    Markdown lines (with method `to_md_lines`)

    Parameters
    ----------
    lines
        list of str, numpydoc.docscrape.Parameter, ...
        What's inside the list depends on the numpydoc section
        but the types are never mixed
    config
        "Global" configuration - parameters that will be used
        in the `core` module and passed around everywhere

    Examples
    --------
    >>> from numpydoc.docscrape import FunctionDoc
    >>> from typing import List
    >>>
    >>> # example function
    >>> def add(a:int, b:int):
    ...     """
    ...     Sum of two integers
    ...     """
    ...     return a + b
    >>>
    >>> # create a parser
    >>> class DummyParser(DocParser):
    ...     def to_md_lines(self) -> List[str]:
    ...         return self.lines
    >>>
    >>> # get lines of numpydoc section "Summary"
    >>> doc = FunctionDoc(add)
    >>> lines = doc['Summary']
    >>>
    >>> # parse section "Summary"
    >>> parser = DummyParser(lines=lines)
    >>> parser.to_md_lines()
    ['Sum of two integers']
    '''
    lines:list
    config:Config = dataclassfield(default_factory=lambda: Config())

    @abstractmethod
    def to_md_lines(self) -> List[str]:
        """
        Implement a method that converts the lines (list of str,
        numpydoc.docscrape.Parameter, ... - what's inside the list
        depends on the numpydoc section) defined in the instance
        to markdown.
        """
        pass  # pragma: no cover


@dataclass(frozen=True)
class PyObjParser(ABC):
    '''
    Abstract class for loading a Python object from a string, extracting information from it
    and then outputing this information to Markdown lines using method `to_md_lines`.

    Parameters
    ----------
    obj_namespace
        String pointing to any Python class, function, method, ... or any object
        that has a docstring e.g. "datetime.datetime", "pandas.Timestamp",
        "numpy.array"
    config
        "Global" configuration - parameters that will be used
        in the `core` module and passed around everywhere

    Examples
    --------
    >>> # create a parser
    >>> class DummyParser(PyObjParser):
    ...     def to_md_lines(self):
    ...         return f'**{self.sig_name}**_{inspect.signature(self.obj)}_'
    >>>
    >>> # extract information from given Python object (imported with a string)
    >>> # and convert it to Markdown
    >>> parser = DummyParser(obj_namespace='numpydoc.docscrape.NumpyDocString')
    >>> parser.to_md_lines()
    '**numpydoc.docscrape.NumpyDocString**_(docstring, config=None)_'
    '''
    obj_namespace:str = dataclassfield(default=None)
    config:Config = dataclassfield(default_factory=lambda: Config())
    sig_name:str = dataclassfield(init=False)
    obj:Any = dataclassfield(init=False)

    def __post_init__(self):
        # get Python object
        obj = pydoc.locate(self.obj_namespace)
        # we handle the unlikely case where the user actually was refering to "None"
        if obj is None and self.obj_namespace not in ('builtins.None', 'None'):
            raise NonExistentObjectException(f'Could not load Python object "{self.obj_namespace}"')

        # get name of Python object for the signature
        alias = self.config.alias
        sig_name = alias if alias is not None else self.obj_namespace
        assert isinstance(sig_name, str)

        # because we are in a frozen dataclass we need this workaround to set attributes
        object.__setattr__(self, 'obj', obj)
        object.__setattr__(self, 'sig_name', sig_name)

    @abstractmethod
    def to_md_lines(self, **kwargs) -> List[str]:
        """
        Implement a method that extracts information from given
        python object `obj` (a module, class, class method, function, coroutine, ...)
        such as the signature and converts it to markdown lines.
        """
        pass  # pragma: no cover


# -

# # Parser for pure text sections such as "summary" and "extended summary"

class TextSectionParser(DocParser):
    """
    Parses sections of a numpydoc docstring that contains only
    text (as opposed to for instance the section "Parameters"
    where numpydoc gives us a list of parameters objects).
    """

    def to_md_lines(self) -> List[str]:
        # precision on the expected types inside the list `self.lines`
        self.lines:List[str]
        return self.lines


# # Parser for "Examples" section

# ## Helper to parse a single block

@dataclass(frozen=True)
class ExampleBlockHandler:
    """
    Helper for ExampleParser, converts a single example block in a docstring
    to Markdown lines (see property `markdown_lines`)
    """
    block:ExampleBlock
    config:Config

    @property
    def output_lang(self):
        # output langs indicated directly in examples
        # e.g. {{python}} will override the parameter
        # `examples_md_lang` (as explained in the wiki of the library)
        if self.block.output_lang is None:
            lang:str = self.config.examples_md_lang
            # there should always be a default lang, this cannot be None or such
            assert isinstance(lang, str)
        else:
            lang:str = self.block.output_lang
        return '' if lang == SpecialExampleOutputLanguages.RAW.value else lang

    @staticmethod
    def _text_line_to_md(example_line:ExampleLine) -> List[str]:
        return [example_line.line]

    @staticmethod
    def _output_lang_line_to_md(example_line:ExampleLine) -> List[str]:
        return []

    def _output_line_to_md(self, example_line:ExampleLine) -> List[str]:
        new_lines = []
        # add markdown syntax for begining of code block
        if example_line.global_index == self.block.ix_first_output:
            if self.output_lang != SpecialExampleOutputLanguages.MARKDOWN_RENDERED.value:
                new_lines.append(f'```{self.output_lang}')

        # clean the output line before adding it
        line = example_line.line
        if self.config.remove_doctest_blanklines:
            line = Patterns.blankline.sub(repl='', string=line)
        new_lines.append(line)

        # add markdown syntax for end of code block
        # (note that if there is a single line of output
        #  we will add both beginning and end sign for markdown
        #  code block)
        if example_line.global_index == self.block.ix_last_output:
            if self.output_lang != SpecialExampleOutputLanguages.MARKDOWN_RENDERED.value:
                new_lines.append('```')

        return new_lines

    def _input_line_to_md(self, example_line:ExampleLine) -> List[str]:
        new_lines = []
        # add markdown syntax for begining of code block
        # (same logic as for method `_output_line_to_md`)
        if example_line.global_index == self.block.ix_first_input:
            new_lines.append('```python')

        # clean and then add line
        line = example_line.line
        if self.config.remove_doctest_skip:
            line = Patterns.doctest_skip.sub(repl='', string=line)
        new_lines.append(Patterns.console_py.sub(repl='', string=line))

        # add markdown syntax for end of code block
        if example_line.global_index == self.block.ix_last_input:
            new_lines.append('```')

        return new_lines

    @property
    def markdown_lines(self) -> List[str]:
        transformers = {ExampleLineType.OUTPUT_LANG:self._output_lang_line_to_md,
                        ExampleLineType.TEXT:self._text_line_to_md,
                        ExampleLineType.INPUT:self._input_line_to_md,
                        ExampleLineType.OUTPUT:self._output_line_to_md}
        new_lines = []
        for example_line in self.block.example_lines:
            # get the appropriate function then pass the example line object
            f = transformers[example_line.line_type]
            new_lines.extend(f(example_line=example_line))
        return new_lines


# ## Parser for the full example section

class ExampleParser(DocParser):
    """
    Parses the "Examples" section of a numpydoc docstring
    """

    def to_md_lines(self) -> List[str]:
        # precision on the expected types inside the list `self.lines`
        self.lines:List[str]
        new_lines = []
        blocks:List[ExampleBlock] = ExamplesLabeller(lines=self.lines).example_blocks
        for block in blocks:
            ebh = ExampleBlockHandler(block=block, config=self.config)
            new_lines.extend(ebh.markdown_lines)
        return new_lines


# # Parser for sections similar to `Parameters`

class ParametersLikeSectionParser(DocParser):
    """
    Parses any section of a numpydoc docstring that "behaves"
    like the "Parameters" section e.g. "Other Parameters",
    "Returns", ...

    For these sections numpydoc gives back a list of parameters
    objects (numpydoc.docscrape.Parameter)
    """

    def _param_to_md_lines(self, param:Parameter) -> List[str]:
        new_lines:List[str] = []

        # get name and type and escape them
        name = MdEscaper.escape(param.name).strip()
        name = f'**{name}**' if name != '' else ''
        type_ = param.type.strip()
        type_ = MdEscaper.escape(type_)
        type_ = f"**_{param.type}_** " if param.type != '' else ''
        prefix = '* ' if name != '' or type_ != '' else ''

        # get separator between name and type
        sep_name_type = ' : ' if (name != '' and type_ != '') else ''

        # construct the first line (* {parameter_name} : {parameter_type})
        new_lines.append(f'{prefix}{name}{sep_name_type}{type_}'.strip())

        # get lines of description
        desc_lines = param.desc
        if not len(desc_lines):
            return new_lines

        # if there is a description separate it with a line break
        new_lines.append('')

        # format lines of description
        new_lines.extend(param_description_lines_to_md_lines(desc_lines))
        return new_lines

    def to_md_lines(self) -> List[str]:
        # precision on the expected types inside the list `self.lines`
        self.lines:List[Parameter]
        nb_lines = len(self.lines)
        new_lines = []
        for ix, param in enumerate(self.lines):
            new_lines.extend(self._param_to_md_lines(param))
            # add a line break between parameters. This does
            # not change the rendering but looks prettier
            # when not rendered
            is_last_param = ix == (nb_lines - 1)
            if not is_last_param:
                new_lines.append('')
        return new_lines


# # Parser for "See Also" section

class SeeAlsoParser(DocParser):
    """
    Parses the "See Also" section of a numpydoc docstring
    """

    def to_md_lines(self) -> List[str]:
        # precision on the expected types inside the list `self.lines`
        self.lines:List[Tuple[list, list]]
        nb_lines = len(self.lines)
        new_lines:List[str] = []

        for ix, (funcs, desc_lines) in enumerate(self.lines):
            # example of (funcs, description):
            # ([('example_func2', None), ('example_func3', None)], ['description'])
            # not sure what None could be (second item of each tuple in funcs) ?!

            # add object names
            names = ', '.join(MdEscaper.escape(func[0]).strip() for func in funcs)
            new_lines.append(f"* **{names}**")

            # get description lines
            desc_lines = param_description_lines_to_md_lines(desc_lines)
            if len(desc_lines):
                # if there is a description separate it with a line break
                new_lines.append('')
                new_lines.extend(desc_lines)

            # like with section "Parameters", add a line break between objects
            is_last_param = ix == (nb_lines - 1)
            if not is_last_param:
                new_lines.append('')

        return new_lines


# # Special parsing for the signature

class SignatureParser(PyObjParser):
    """
    Parses the signature of a Python object
    """

    @staticmethod
    def get_signature(obj:Any) -> str:
        """
        Gets the signature of given object as a string.
        This handles the case when the object `obj` is a builtin.

        inspect.signature(self.obj) fails with some builtin objects
        because they are written in C and they cannot be introspected.

        Notes
        -----
        I made this as a staticmethod so it's easier to test with doctest.

        Examples
        --------
        >>> SignatureParser.get_signature(bool)
        '(self, /, *args, **kwargs)'

        >>> from npdoc_to_md import render_string
        >>> SignatureParser.get_signature(render_string)
        '(string: str, ignore_errors: bool = False) -> str'
        """
        assert callable(obj), f'Trying to get signature of non callable object: {obj}'

        try:
            # this will fail for builtins written in C but the except part will handle that
            return str(inspect.signature(obj))
        except ValueError as e:
            if hasattr(obj, '__init__'):
                return str(inspect.signature(obj.__init__))
            else:
                raise SignatureNotFoundException(f'Cannot get signature of object {obj}') from e  # pragma: no cover

    def to_md_lines(self) -> List[str]:
        name = MdEscaper.escape(self.sig_name)
        name_colored = f'<span style="color:purple">{name}</span>'

        # get the level (title, subtitle etc.)
        level = self.config.md_section_level - 1 if self.config.md_section_level > 1 else 1

        # handle case of non callable (e.g. property)
        if not callable(self.obj):
            return [f'{"#" * level} {name_colored}']

        # get signature string and "clean" it (remove self or cls argument)
        sig_str = self.get_signature(obj=self.obj)
        sig_str = MdEscaper.escape(Patterns.self_or_cls.sub(repl='', string=sig_str))
        return [f'{"#" * level} {name_colored}_{sig_str}_']


# # Class to parse every section

# +
class DefaultParsers:
    """
    Parsers for the different sections described in numpydoc style docstrings.

    Attributes
    ----------
    sections : dict[str, DocScraper|PyObjScraper]
        A dictionary with a parser class (values of the dict)
        for each numpydoc section (keys of the dict)
    """
    _parameter_like_sections = ('Returns', 'Yields', 'Parameters', 'Attributes', 'Methods',
                                'Raises', 'Warns', 'Receives', 'Other Parameters')
    _text_sections = ('Summary', 'Extended Summary', 'Notes', 'Warnings', 'References')
    sections = {'Examples':ExampleParser, 'Signature':SignatureParser, 'See Also':SeeAlsoParser}
    sections.update({k:ParametersLikeSectionParser for k in _parameter_like_sections})
    sections.update({k:TextSectionParser for k in _text_sections})


class ParserEngine(PyObjParser):
    """
    Uses the different parsers defined in module npdoc_to_md.parser to convert
    a numpydoc docstring to Markdown
    """

    def section_to_md_lines(self, sections_finder:SectionsFinder, section:str) -> List[str]:
        """
        Converts any `section` of a numpydoc style docstring to markdown lines
        """
        # skip empty sections (except for "Signature" which is done using `inspect.signature`)
        lines = sections_finder[section]
        if section != 'Signature' and not len(lines):
            return []

        # handle bug where a section does not exists and it is still parsed by numpydoc
        # for instance this happened with the section Attributes of pd.DataFrame
        if section not in sections_finder.sections_without_headers and section not in sections_finder.all_visible_sections:
            return []

        # get and instantiate parser class
        parser_class = DefaultParsers.sections.get(section, TextSectionParser)
        if issubclass(parser_class, DocParser):
            parser = parser_class(lines=lines, config=self.config)
        elif issubclass(parser_class, PyObjParser):
            parser = parser_class(obj_namespace=self.obj_namespace, config=self.config)
        else:  # pragma: no cover
            raise TypeError(f'Parser obtained for numpy section "{section}" does not '
                            'subclasss npdoc_to_md.parser.DocParser or '
                            'npdoc_to_md.parser.PyObjParser')
        return parser.to_md_lines()

    def to_md_lines(self) -> List[str]:
        # initialize tool to find sections
        sections_finder = SectionsFinder(self.obj, ignore_custom_section_warning=self.config.ignore_custom_section_warning)
        custom_section_names = list(sections_finder.custom_sections.keys())
        if len(custom_section_names):
            log(f'Found the following custom sections {custom_section_names} in object {self.obj}.\n'
                '(we will interpret these custom sections as markdown text like sections '
                '"Summary" and "Extended Summary")')

        # convert each section in order of appearance to markdown
        new_lines = []
        for section in sections_finder.all_sections:
            new_lines_sublist = self.section_to_md_lines(sections_finder=sections_finder, section=section)
            # after converting if we end up with an empty list we'll skip it
            if not len(new_lines_sublist):
                continue

            # add a header and two line breaks (only one is needed but this is
            # prettier for the non rendered version since you'll have a separation line
            # below the section)
            if section not in sections_finder.sections_without_headers:
                new_lines.append(f'{self.config.md_section_level*"#"} {section}\n\n')

            # add the new lines
            new_lines.append('\n'.join(new_lines_sublist) + '\n\n')
        return new_lines

    def to_md_string(self) -> str:
        """
        Creates a single Markdown string using the output of method
        `to_md_lines`
        """
        return ''.join(self.to_md_lines()).strip()


# -

# # Main function

def parse_and_render(obj_namespace:str, config:Config) -> str:
    """
    Loads object at given namespace (`obj_namespace`)
    e.g. `pandas.DataFrame` and using given configuration
    transforms its docstring to Markdown (optionally
    for desginated members of this object too).
    """
    # case of a single item to parse
    engine = ParserEngine(obj_namespace=obj_namespace, config=config)
    result = engine.to_md_string()
    if config.members is None or len(config.members) == 0:
        return result

    # parse members
    results = [result]
    parsed_members = MembersParser(obj=engine.obj, members=config.members).parsed_members
    for member in parsed_members:

        # we have to make a few changes to the config for members
        config_dict = dataclasses.asdict(config)
        config_dict.pop('members')

        if config_dict['alias'] is not None:
            config_dict['alias'] = f"{config_dict['alias']}.{member}"
        new_config = Config(**config_dict)
        results.append(ParserEngine(obj_namespace=f'{obj_namespace}.{member}', config=new_config).to_md_string())

    return '\n\n'.join(results)



"""
Tests the sections module
"""
import pytest
from npdoc_to_md.sections import SectionsFinder


# # Tests

# I am putting the lines on top of each other in the tuples on purpose
# so we can see them better, sorry for the ugly display
@pytest.mark.parametrize('line, next_line, expected', [('My Section',  # "normal" case
                                                        '----------',
                                                        'My Section'),
                                                       # with leading spaces (should still work)
                                                       ('  My Section',
                                                        '  ----------',
                                                        'My Section'),
                                                       # not enough hyphens (for this test,
                                                       # the length of both lines must match
                                                       # after being rstripped)
                                                       ('My Section',
                                                        ' ---------',
                                                        None),
                                                       # too many hyphens
                                                       (' My Section',
                                                        '-----------',
                                                        None),
                                                       # something that looks like a section
                                                       # but is not one
                                                       ('  Not a section',
                                                        '  -----foo-----',
                                                        None)])
def test_finding_sections_from_lines(line, next_line, expected):
    assert SectionsFinder._find_from_lines(line=line, next_line=next_line) == expected



# +
"""
Tests the core module. Basically end to end tests.
"""
import os
import pytest
import tempfile
from dataclasses import dataclass
from typing import Any, List, Optional

# local imports
from npdoc_to_md.core import RenderedFile, render_obj_docstring, render_string, render_file, render_folder
from npdoc_to_md.config import MemberFlag
from npdoc_to_md.tests.conftest import CLIRunner, CustomNamedTemporaryFile
from npdoc_to_md.tests.expectations import (builtins_none_md,
                                            documented_func_example_md,
                                            documented_generator_func_example_md,
                                            DocumentedClassExampleMd,
                                            empty_function_md,
                                            EmptyClassMd,
                                            EmptyClassDictMd,
                                            empty_doc_sections_function_md,
                                            EmptyDocSectionsClass_md,
                                            none_md,
                                            testing_module_md,
                                            RENDERING_OPTIONS)


# -

# # Helpers

# ## Generating placeholders

class PlaceholderStringGenerator:
    """
    Creates a placeholder string (e.g. for putting into a template file)
    from given placeholder parameters.

    Examples
    --------
    >>> PlaceholderStringGenerator.generate(obj="pandas.DataFrame",
    ...                                     alias="pd.DataFrame",
    ...                                     examples_md_lang='raw',
    ...                                     remove_doctest_skip=True,
    ...                                     remove_doctest_blanklines=True,
    ...                                     md_section_level=3,
    ...                                     ignore_custom_section_warning=True,
    ...                                     members=['public$', '+__dict__', '-to_dict'])  # doctest: +SKIP
    "{{'obj':'pandas.DataFrame', 'alias':'pd.DataFrame', 'examples_md_lang':'raw', 'remove_doctest_skip':true, \
'remove_doctest_blanklines':true, 'md_section_level':'3', 'ignore_custom_section_warning':true, \
'members':['public$', '+__dict__', '-to_dict']}}"
    """

    def _format_value(param_name:str, param_value:Any):
        if param_name in ('remove_doctest_blanklines', 'remove_doctest_skip', 'ignore_custom_section_warning'):
            assert isinstance(param_value, bool)
            return str(param_value).lower()  # true, false

        elif param_name == 'members':
            assert isinstance(param_value, list)
            return '[' + ','.join(f'"{s}"' for s in param_value) + ']'

        else:
            return f'"{param_value}"'

    def generate(**placeholder_params):
        args = []
        for k, v in placeholder_params.items():
            v = PlaceholderStringGenerator._format_value(param_name=k, param_value=v)
            args.append(f'"{k}":{v}')
        return '{{' + ','.join(args) + '}}'


# ## "Flexible" tester for pytest

@dataclass(frozen=True)
class PytestTester:
    """
    Tool for testing the rendering of the docstring of a Python object
    using the various functions of the library in Python and in the CLI:
    * `render_obj_docstring`: we can pass the object (actually its namespace) directy to this function
    * `render_string`: we will pass the object  (actually its namespace) through a placeholder
    * `render_file`: we will create a temporary file with only a placeholder pointing to
      the object
    * `render_folder`: we will create a temporary directory and there add a file with only
      a placeholder pointing to the object

    Notes
    -----
    This could probably by DRYer...

    Parameters
    ----------
    obj_namespace
        Importable string e.g. "pandas.DataFrame".
        In `render_obj_docstring` as in placeholders this is referred to as parameter `obj`
    expected
        Expected Markdown string after rendering the docstring
    rendering_options
        See options in docstring of `render_obj_docstring`
    use_cli
        If False uses Python otherwise uses the CLI
    hyphenize
        Only relevant for the CLI. If True underscores are converted to hyphens.
        This is for testing whether underscores and hyphens can be switched
        in commands and parameters
    """
    obj_namespace:str
    expected:str
    rendering_options:dict
    use_cli:bool
    hyphenize:bool

    def _run_in_cli(self, func, kwargs, extras:Optional[List[str]]=None):
        extras = [] if extras is None else extras
        result = CLIRunner(func=func, kwargs=kwargs, command=['npdoc_to_md', func.__name__],
                           hyphenize=self.hyphenize).run(extras=extras).strip()
        # IMPORANT! The line below will normalize line breaks to \n (in Windows it would be \r\n)
        return '\n'.join(result.splitlines())

    def using_obj(self):
        # prepare function and params
        func = render_obj_docstring
        kwargs = self.rendering_options.copy()
        if self.use_cli:
            # we need to escape None for the CLI otherwise it will be interpreted as
            # a None instead of the string "None"
            if self.obj_namespace == 'None':
                kwargs['obj'] = "'None'"
            else:
                kwargs['obj'] = self.obj_namespace
            if 'members' in kwargs:
                kwargs['members'] = str(kwargs['members'])
        else:
            kwargs['obj'] = self.obj_namespace

        # get and check result
        result = self._run_in_cli(func=func, kwargs=kwargs) if self.use_cli else func(**kwargs)
        assert result == self.expected

    def using_string(self):
        # prepare function and params
        func = render_string
        placeholder_string = PlaceholderStringGenerator.generate(obj=self.obj_namespace, **self.rendering_options)

        # get result
        # for testing the function "npdoc_to_md.render_string" with the CLI
        # we do a little "hack" where we prepend a dummy line
        # to avoid awkward escaping (multiline strings are treated differently)...
        if self.use_cli:
            string = f'Demo:\n{placeholder_string}'
            expected = f'Demo:\n{self.expected}'
            result = self._run_in_cli(func=func, kwargs={'string':string})
        else:
            result = func(string=placeholder_string)
            expected = self.expected
        assert result == expected

    def using_file(self):
        # preparations
        func = render_file
        placeholder_string = PlaceholderStringGenerator.generate(obj=self.obj_namespace, **self.rendering_options)

        # work in temp files
        with CustomNamedTemporaryFile() as input_tmp, CustomNamedTemporaryFile() as output_tmp:

            # write input file
            source = input_tmp.name + '.npmd'
            destination = output_tmp.name + '.md'
            with open(source, mode='w', encoding='utf-8', newline='\n') as fh:
                fh.write(placeholder_string)

            # get result
            kwargs = dict(source=source, destination=destination)
            if self.use_cli:
                # the extras will allow us to access the property "rendered_text"
                # of the class RenderedFileCLI (the python-fire library we use for the CLI has such features)
                result = self._run_in_cli(func=func, kwargs=kwargs, extras=['-', 'rendered_text'])
            else:
                rendered_file:RenderedFile = func(**kwargs)
                result = rendered_file.rendered_text
            assert result == self.expected

            # check output file as well
            with open(destination, mode='r', encoding='utf-8', newline='\n') as fh:
                file_result = fh.read()
            assert file_result == self.expected

    def using_folder(self):
        # preparations
        func = render_folder
        placeholder_string = PlaceholderStringGenerator.generate(obj=self.obj_namespace, **self.rendering_options)

        # work in temp dir
        with tempfile.TemporaryDirectory() as inputdirpath, tempfile.TemporaryDirectory() as outputdirpath:

            # write a test file
            test_filepath = os.path.join(inputdirpath, 'Test.npmd')
            with open(test_filepath, mode='w', encoding='utf-8', newline='\n') as fh:
                fh.write(placeholder_string)

            # get result
            kwargs = dict(source=inputdirpath, destination=outputdirpath)
            if self.use_cli:
                # the extras will allow us to access the property "rendered_text" of the first item
                # (an instance of the class RenderedFileCLI)
                result = self._run_in_cli(func=func, kwargs=kwargs, extras=['-', '0', '-', 'rendered_text'])
            else:
                rendered_files:List[RenderedFile] = func(**kwargs)
                result = rendered_files[0].rendered_text
            assert result == self.expected

            # check output file as well
            test_filepath_converted = os.path.join(outputdirpath, 'Test.md')
            with open(test_filepath_converted, mode='r', encoding='utf-8', newline='\n') as fh:
                file_result = fh.read()
            assert file_result == self.expected


# ## Expectations as parameters for pytest

# [(obj_namespace, expected_rendered_docstring)]
doc_expectations = [('npdoc_to_md.testing.documented_func_example', documented_func_example_md),
                    ('npdoc_to_md.testing.documented_generator_func_example', documented_generator_func_example_md),
                    ('npdoc_to_md.testing.DocumentedClassExample', DocumentedClassExampleMd),
                    ('npdoc_to_md.testing', testing_module_md),
                    # objects with empty docstrings or sections
                    ('npdoc_to_md.testing.empty_function', empty_function_md),
                    ('npdoc_to_md.testing.EmptyClass', EmptyClassMd),
                    ('npdoc_to_md.testing.EmptyClass.__dict__', EmptyClassDictMd),
                    ('npdoc_to_md.testing.empty_doc_sections_function', empty_doc_sections_function_md),
                    ('npdoc_to_md.testing.EmptyDocSectionsClass', EmptyDocSectionsClass_md),
                    ('None', none_md),
                    ('builtins.None', builtins_none_md)]


# # Test expectations of rendering for our tests functions

@pytest.mark.parametrize('test_method', [m for m in dir(PytestTester) if m.startswith('using_')])
@pytest.mark.parametrize('runner', ['cli_underscore', 'cli_hyphen', 'python'])
@pytest.mark.parametrize('obj_namespace, expected', doc_expectations,
                         # don't show the long docstrings in the pytests ids
                         ids=[item[0] for item in doc_expectations])
def test_render_docstring(test_method, runner, obj_namespace, expected):

    # handle special case where we want to test the rendering of the docstrings of class members
    rendering_options = RENDERING_OPTIONS.copy()
    if obj_namespace == 'npdoc_to_md.testing.DocumentedClassExample':
        rendering_options['members'] = [MemberFlag.PUBLIC.value]

    pt = PytestTester(obj_namespace=obj_namespace, expected=expected, rendering_options=rendering_options,
                      use_cli=runner.startswith('cli'), hyphenize=runner.endswith('hyphen'))
    getattr(pt, test_method)()



# +
"""
Configurations and helpers for test modules
"""
import inspect
import logging
import os
import tempfile
from dataclasses import dataclass, field as dataclassfield
from typing import Callable, List, Optional, Union

from npdoc_to_md.logger import log


# -

# # Helpers

# ## CLI Runner

@dataclass(frozen=True)
class CLIRunner:
    """
    Runs a Python function in the CLI (see `run` method).

    The attributes `mandatory`, `optionals` and `flags` (see below)
    are infered from `kwargs` (see Parameters) and the signature of `func`.

    Parameters
    ----------
    func
    kwargs
        All parameters you would pass to the function if it were
        in Python and all given as keywords arguments.
        This means that positional only parameters will not work.
    command
        Command corresponding to the `func` in the CLI
        or list of command + subcommand(s)
    hyphenize
        If True, converts underscores to hyphens
        (used for testing if those characters are
         interchangeable as expected in our CLI)

    Attributes
    ----------
    mandatory
        Mandatory arguments in the form of a dict
    optionals
        Optional arguments in the form of a dict
    flags
        CLI flags e.g. "--verbose"

    Examples
    --------
    >>> from npdoc_to_md.core import render_obj_docstring
    >>>
    >>> cli = CLIRunner(func=render_obj_docstring,
    ...                 command=['npdoc_to_md', 'render_obj_docstring'],
    ...                 kwargs=dict(obj='npdoc_to_md.testing.now_utc',
    ...                             alias="now_utc",
    ...                             examples_md_lang='raw',
    ...                             remove_doctest_skip=True,
    ...                             remove_doctest_blanklines=True,
    ...                             md_section_level=3),
    ...                 hyphenize=True)
    >>> cli
    CLIRunner(command=['npdoc-to-md', 'render-obj-docstring'], mandatory={'obj': 'npdoc_to_md.testing.now_utc'}, \
optionals={'alias': 'now_utc', 'examples-md-lang': 'raw', 'md-section-level': 3}, \
flags=['remove-doctest-blanklines', 'remove-doctest-skip'])
    """
    func:Callable = dataclassfield(repr=False)
    kwargs:dict = dataclassfield(repr=False)
    command:Union[str, List[str]]
    hyphenize:bool = dataclassfield(default=False, repr=False)
    mandatory:dict = dataclassfield(init=False)
    optionals:dict = dataclassfield(init=False)
    flags:List[str] = dataclassfield(init=False)

    def __post_init__(self):
        # verifications
        assert isinstance(self.command, (list, str))
        assert len(self.command) > 0

        # inspect parameters of `func`
        params:List[inspect.Parameter] = inspect.signature(self.func).parameters
        mandatory, optionals, flags = {}, {}, []
        for param in params.values():

            # handle non provide parameters
            # raise if mandatory param missing
            is_mandatory = param.default is inspect._empty
            if param.name not in self.kwargs:
                if is_mandatory:  # pragma: no cover
                    raise ValueError(f'Missing mandatory parameter {param.name}')
                continue

            # classify parameters
            if is_mandatory:
                mandatory[param.name] = self.kwargs[param.name]
            elif isinstance(param.default, bool):
                value = self.kwargs[param.name]
                if not isinstance(value, bool):
                    raise TypeError(f'Expected bool for parameter "{param.name}". Got type {type(value)}')
                # flags are only used to denote a True value
                if value is True:
                    flags.append(param.name)
            else:
                optionals[param.name] = self.kwargs[param.name]

        # convert underscores to hyphens
        if self.hyphenize:
            rename = lambda k: k.replace('_', '-')
            # rename command
            # note: most likely we will always provide commands
            # as a list (the library works with subcommands) so no need to cover this
            if isinstance(self.command, str):  # pragma: no cover
                command = rename(self.command)
            else:
                assert isinstance(self.command, list)
                command = [rename(c) for c in self.command]
            object.__setattr__(self, 'command', command)

            # rename params
            mandatory = {rename(k):v for k, v in mandatory.items()}
            optionals = {rename(k):v for k, v in optionals.items()}
            flags = [rename(v) for v in flags]

        object.__setattr__(self, 'mandatory', mandatory)
        object.__setattr__(self, 'optionals', optionals)
        object.__setattr__(self, 'flags', flags)

    def run(self, extras:Optional[List[str]]=None):
        """
        Runs our command in the CLI

        Parameters
        ----------
        extras
            Any extra arguments you want. They will be put at the end.
            This is useful for Python-fire if we want to access
            properties.
            E.g. "npdoc-to-md render-folder . - 0 - rendered_text"
            will give us the attribute "rendered_text" of the first rendered file
        """
        # get plumbum
        try:
            from plumbum import local
        except ModuleNotFoundError as e:  # pragma: no cover
            raise ModuleNotFoundError('Please install testing dependency `plumbum` (pip install plumbum)') from e

        # get command
        commands = [self.command] if isinstance(self.command, str) else self.command
        command, args = commands[0], commands[1:]
        runner = local[command]

        # get params
        extras = [] if extras is None else extras
        for k, v in self.mandatory.items():
            args.append(f'-{k}')
            args.append(v)
        for k, v in self.optionals.items():
            args.append(f'--{k}')
            args.append(v)
        args.extend([f'--{f}' for f in self.flags])
        args.extend([e.replace('_', '-') if self.hyphenize else e for e in extras])

        log(f'Running command {command} with args: {args}', level=logging.INFO)

        return runner.__getitem__(args)()  # cannot do something like runner[*args]()


# ## Temporary files

class CustomNamedTemporaryFile:
    """
    This custom implementation is needed because of the following limitation of tempfile.NamedTemporaryFile:

    > Whether the name can be used to open the file a second time, while the named temporary file is still open,
    > varies across platforms (it can be so used on Unix; it cannot on Windows NT or later).

    See https://stackoverflow.com/a/63173312
    """

    def __init__(self, mode='wb', delete=True):
        self._mode = mode
        self._delete = delete

    def __enter__(self):
        # Generate a random temporary file name
        file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        # Ensure the file is created
        open(file_name, "x").close()
        # Open the file in the given mode
        self.file = open(file_name, self._mode)
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()
        if self._delete:
            os.remove(self.file.name)



# +
"""
Labels lines of docstrings (right now used for examples
to distinguish between lines of inputs/ouputs or "text"
lines outside of examples)
"""
from dataclasses import dataclass
from enum import IntEnum, auto as enumauto
from typing import List, Union

# local imports
from npdoc_to_md.helpers import Patterns, previous_current_next


# -

# # Labels

class ExampleLineType(IntEnum):
    """
    Labels for lines in the example section of a docstring

    Attributes
    ----------
    INPUT : int
        Inputs found in examples (>>> or ...)
    OUTPUT : int
        Outputs of examples
    TEXT : int
        Anything before inputs that is not part of the output of a previous example
    OUTPUT_LANG : int
        Special placeholder of this library that tells us
        how to represent outputs (as markdown, python code, raw code, ...)
    """
    INPUT = enumauto()
    OUTPUT = enumauto()
    TEXT = enumauto()
    OUTPUT_LANG = enumauto()


# # Helper dataclasses

# +
@dataclass(frozen=True)
class ExampleLine:
    global_index:int
    line:str
    line_type:ExampleLineType


@dataclass(frozen=True)
class ExampleBlock:
    """
    Represents a single example block which may contain:
    * placeholders for the language of the output of the example
    * text
    * inputs
    * output

    For instance:

    {{python}}
    Some random example
    >>> a = 'foo'
    >>> a
    'foo'
    """

    example_lines:List[ExampleLine]

    @property
    def output_lang(self) -> str:
        # locate the first line indicating the language for the example output e.g. {{html}}
        iterator = (exl.line for exl in self.example_lines if exl.line_type is ExampleLineType.OUTPUT_LANG)
        flav = next(iterator, None)
        return flav.lstrip('{').rstrip('}').strip() if flav is not None else None

    def _get_ix(self, operator:Union[min, max], line_type:ExampleLineType):
        """
        Helper to make properties for notable line indices in the example block
        e.g. property `ix_first_input`.
        """
        iterator = (exl.global_index for exl in self.example_lines if exl.line_type is line_type)
        return operator(iterator, default=None)

    @property
    def ix_first_input(self) -> int:
        return self._get_ix(operator=min, line_type=ExampleLineType.INPUT)

    @property
    def ix_last_input(self) -> int:
        return self._get_ix(operator=max, line_type=ExampleLineType.INPUT)

    @property
    def ix_first_output(self) -> int:
        return self._get_ix(operator=min, line_type=ExampleLineType.OUTPUT)

    @property
    def ix_last_output(self) -> int:
        return self._get_ix(operator=max, line_type=ExampleLineType.OUTPUT)

    def __repr__(self):
        # for once we should overwrite the dataclass repr
        # this will look much better
        tab = ' ' * 4  # don't use \t otherwise this will be painful with doctest
        exl_str = '\n'.join(f'{tab}{repr(exl)}' for exl in self.example_lines)
        return f'{self.__class__.__name__}\n{exl_str}'


# -

# # Main object

@dataclass(frozen=True)
class ExamplesLabeller:
    """
    Puts a label for each line in the example section of a docstring.
    The labels distinguish inputs (>>>, ...), ouputs and any text
    that preceedes inputs.

    Examples
    --------
    >>> from pprint import pprint
    >>>
    >>> lines = ['{{python}}',
    ...          '* first example',
    ...          '>>> print(123)',
    ...          '123',
    ...          '',
    ...          '* second example',
    ...          '>>> print("hello")',
    ...          'hello']
    >>> el = ExamplesLabeller(lines=lines)
    >>> pprint(el.example_blocks)
    [ExampleBlock
        ExampleLine(global_index=0, line='{{python}}', line_type=<ExampleLineType.OUTPUT_LANG: 4>)
        ExampleLine(global_index=1, line='* first example', line_type=<ExampleLineType.TEXT: 3>)
        ExampleLine(global_index=2, line='>>> print(123)', line_type=<ExampleLineType.INPUT: 1>)
        ExampleLine(global_index=3, line='123', line_type=<ExampleLineType.OUTPUT: 2>),
     ExampleBlock
        ExampleLine(global_index=4, line='', line_type=<ExampleLineType.TEXT: 3>)
        ExampleLine(global_index=5, line='* second example', line_type=<ExampleLineType.TEXT: 3>)
        ExampleLine(global_index=6, line='>>> print("hello")', line_type=<ExampleLineType.INPUT: 1>)
        ExampleLine(global_index=7, line='hello', line_type=<ExampleLineType.OUTPUT: 2>)]

    >>> el.example_blocks[0].output_lang
    'python'

    >>> print(el.example_blocks[1].output_lang)
    None
    """
    lines:List[str]

    @property
    def labels(self):
        lines = self.lines
        if len(lines) == 0:  # pragma: no cover
            return []
        labelled_lines = []
        for ix, (previous_line, line, current_line) in enumerate(previous_current_next(lines)):
            # easy case where we know it's an input
            is_input = Patterns.console_py.search(line)
            if is_input:
                labelled_lines.append(ExampleLine(global_index=ix, line=line, line_type=ExampleLineType.INPUT))
                continue

            # distinguish outputs and comments before tests
            previous_label = labelled_lines[ix - 1].line_type if ix > 0 else None
            # note that an empty line ends an example
            # indeed, in pytest you even have to use <BLANKLINE> to indicate an empty line as part of an output
            follows_input = previous_line is not None and previous_label is ExampleLineType.INPUT
            follows_output = previous_label is ExampleLineType.OUTPUT
            is_empty_line = line.strip() == ''
            previous_line_empty = previous_line is not None and previous_line.strip() == ''
            is_example_output = ((follows_input and not is_input and not is_empty_line) or
                                 (follows_output and not is_empty_line and not previous_line_empty))
            if is_example_output:
                labelled_lines.append(ExampleLine(global_index=ix, line=line, line_type=ExampleLineType.OUTPUT))
            else:
                # if not an example output we have to assume by elimination that it is some information (text)
                # before the example e.g. "* first example"
                stripped = line.strip()
                is_lang = stripped.startswith('{{') and stripped.endswith('}}')
                line_type = ExampleLineType.OUTPUT_LANG if is_lang else ExampleLineType.TEXT
                labelled_lines.append(ExampleLine(global_index=ix, line=line, line_type=line_type))
        assert len(labelled_lines) == len(self.lines)
        return labelled_lines

    @property
    def example_blocks(self):
        # find the start of each example block using the labels
        blocks_start = []
        labels = self.labels
        for ix, example_line in enumerate(labels):
            # easy case of the first example
            if ix == 0:
                blocks_start.append(ix)
                continue

            # easy case of examples separated by a line break
            previous = labels[ix - 1]
            if previous.line_type is ExampleLineType.INPUT and example_line.line.strip() == '':
                blocks_start.append(ix)
                continue

            # for the other examples wait for an output line (could also be an empty
            # line separating the examples, I have not made a distinction for that)
            # before
            # declaring a new block
            if previous.line_type is not ExampleLineType.OUTPUT:
                continue
            if example_line.line_type in (ExampleLineType.INPUT, ExampleLineType.TEXT):
                blocks_start.append(ix)

        # get the corresponding line objects
        blocks = []
        for _, start_ix, next_ix in previous_current_next(blocks_start):
            if next_ix is not None:
                blocks.append(ExampleBlock(labels[start_ix:next_ix]))
            # case of last item
            else:
                blocks.append(ExampleBlock(labels[start_ix:]))
        return blocks



"""
Automatic documentation tool using Markdown and Python docstrings written in numpydoc style

See documentation in the wiki folder at the root of the repo or go to https://github.com/ThibTrip/npdoc_to_md/wiki
"""
from npdoc_to_md._version import __version__
from npdoc_to_md.core import RenderedFile, RenderedFileCLI, RenderedFilesCLI, render_obj_docstring, render_string, render_file, render_folder


# +
import inspect
import warnings
from dataclasses import dataclass, field as dataclassfield
from typing import Any, ClassVar, Dict, List, Tuple, Union
from numpydoc.docscrape import ClassDoc, FunctionDoc, NumpyDocString

# local imports
from npdoc_to_md.helpers import previous_current_next


# -

# # Helpers

@dataclass(frozen=True)
class SectionsFinder:
    '''
    Gets all numpydoc style sections in a Python object's docstring
    **including custom defined sections** (numpydoc ignores them :/)
    and in order

    Attributes
    ----------
    sections_without_headers
        Sections of numpydoc for which there is no header e.g. "Signature"
        This is not used in this class by it is used by other modules
    standard_section_names
        Standard sections of numpydoc
    doc
        Object from numpydoc representing a parsed docstring
    docstring_cleaned
        Docstring cleaned with inspect.cleandoc

    Examples
    --------
    >>> def foo(a:int):
    ...     """
    ...     A dummy function
    ...
    ...     Parameters
    ...     ----------
    ...     a
    ...         Some number
    ...
    ...     My custom section
    ...     -----------------
    ...     It works!
    ...     """
    ...     pass
    >>>
    >>> sf = SectionsFinder(foo, ignore_custom_section_warning=True)
    >>> print(sf.all_visible_sections_start_indices)
    {'Parameters': 2, 'My custom section': 7}

    >>> print(sf.all_visible_sections_ranges)
    {'Parameters': range(4, 6), 'My custom section': range(9, 9)}

    >>> print(sf.custom_sections)
    {'My custom section': ['It works!']}

    >>> # note: I don't use pprint because in Python 3.7 we cannot avoid dict sorting
    >>> for k, v in sf.all_sections.items():
    ...     print(f'{k}={repr(v)}')
    Signature=''
    Summary=['A dummy function']
    Extended Summary=[]
    Parameters=[Parameter(name='a', type='', desc=['Some number'])]
    My custom section=['It works!']
    Returns=[]
    Yields=[]
    Receives=[]
    Raises=[]
    Warns=[]
    Other Parameters=[]
    Attributes=[]
    Methods=[]
    See Also=[]
    Notes=[]
    Warnings=[]
    References=''
    Examples=''
    '''
    py_obj:Any
    ignore_custom_section_warning:bool = dataclassfield(default=False, repr=False)
    sections_without_headers:ClassVar[Tuple[str]] = ('Signature', 'Summary', 'Extended Summary')
    # get a list of standard sections
    # note that "index" is not a section but a method listed under NumpyDocString.sections
    standard_section_names:ClassVar[Tuple[str]] = tuple([k for k in NumpyDocString.sections if k != 'index'])
    doc:Union[ClassDoc, FunctionDoc, NumpyDocString] = dataclassfield(init=False)
    docstring_cleaned:str = dataclassfield(init=False)

    def __post_init__(self):

        # pyobj must be a class, function or anything else that has a docstring...
        py_obj = self.py_obj
        if self.ignore_custom_section_warning:
            with warnings.catch_warnings():
                warnings.filterwarnings(action='ignore', category=UserWarning, message='Unknown section')
                doc = self._get_numpydoc_obj(py_obj)
        else:
            doc = self._get_numpydoc_obj(py_obj)

        # clean the docstring, handle the case when there is no docstring (__doc__ is None)
        docstring = py_obj.__doc__
        docstring_cleaned = inspect.cleandoc(docstring) if docstring is not None else ''

        # because we are in a frozen dataclass we need this workaround to set attributes
        object.__setattr__(self, 'doc', doc)
        object.__setattr__(self, 'docstring_cleaned', docstring_cleaned)

    @staticmethod
    def _get_numpydoc_obj(py_obj:Any):
        # special case where __doc__ is None e.g. with non overwritten dunder methods
        # such as "__dict__"
        if hasattr(py_obj, '__doc__') and getattr(py_obj, '__doc__') is None:
            return NumpyDocString('')

        # "normal" cases
        if inspect.isclass(py_obj):
            doc = ClassDoc(py_obj)
        elif inspect.isfunction(py_obj) or inspect.ismethod(py_obj):
            doc = FunctionDoc(py_obj)
        elif hasattr(py_obj, '__doc__'):
            doc = NumpyDocString(py_obj.__doc__)
        else:  # pragma: no cover
            raise TypeError(f'The object {py_obj} is not a class, function, method '
                            'or any other Python object that has a __doc__ attribute')
        return doc

    @property
    def docstring_lines(self):
        return self.docstring_cleaned.splitlines()

    @staticmethod
    def _find_from_lines(line:str, next_line:Union[str, None]) -> Union[str, None]:
        """
        Finds a section in numpydoc style docstring using a given line and the next one
        (or None if it is the last line in the docstring).

        If the line does not seem to correspond to a section, we return None.

        Examples
        --------
        >>> l1 = 'My Section'
        >>> l2 = '----------'
        >>> SectionsFinder._find_from_lines(line=l1, next_line=l2)
        'My Section'
        """
        # skip empty lines
        if next_line is None:  # last line -> cannot be a section
            return None
        if not len(line.strip()) or not len(next_line.strip()):
            return None

        # do an rstrip, we still need leading spaces on the left
        # to handle cases where the section separator is misaligned
        line, next_line = line.rstrip(), next_line.rstrip()

        # check if first letter is uppercase (otherwise we consider that's not a section)
        if not line.strip()[0].isupper():
            return None

        # if the length is not the same, then there is necessarily a misalignement
        # because we did a rstrip before
        # (and thus no section would be detected by numpydoc and we imitate this behavior)
        if len(line) != len(next_line):
            return None
        nb_chars = len(line)

        # go through previous and current line characters to determine if it's a section
        # note that we do need this complex code because like numpydoc we should
        # be able to parse sections that have leading spaces even after using cleandoc()
        # on the docstring
        header_reached = False
        section_name_chars = []
        for ix, (pc, c) in enumerate(zip(line.rstrip(), next_line.rstrip())):
            # check if we reached the header
            if not header_reached:
                header_reached = pc != ' '

            # until we reach the header we expect to have spaces on both lines
            if not header_reached:
                if not (pc == c == ' '):
                    return None
            else:
                section_name_chars.append(pc)
                if c not in ('-', '='):
                    return None

            # if the for loop went through all characters without breaking
            # it means we found a section
            if ix == (nb_chars - 1):
                return ''.join(section_name_chars)

    @property
    def all_visible_sections_start_indices(self) -> Dict[int, str]:
        """
        Returns a dictionary of visible section names and where they
        start (see class Examples)
        """
        sections = {}
        lines = self.docstring_lines
        # case of empty docstring
        if not len(lines):
            return {}
        # normal case
        for ix, (_, line, next_line) in enumerate(previous_current_next(lines)):
            section_name:Union[str, None] = self._find_from_lines(line=line, next_line=next_line)
            if section_name is not None:
                sections[section_name] = ix
        return sections

    @property
    def all_visible_sections_ranges(self) -> Dict[str, range]:
        """
        Returns a dictionary of visible section names and their span
        (see class Examples)
        """
        # Example for the values of variables below:
        # start_indices = {5:'Parameters', 10:'Examples'}
        # shifted = [10, 25]
        # ranges = [range(5, 9), range(10, 25)]
        # return {'Parameters':range(5, 9), 'Examples':range(10, 25)}
        sections_indices = self.all_visible_sections_start_indices
        if not len(sections_indices):
            return {}
        shifted = list(sections_indices.values())[1:] + [len(self.docstring_lines)]
        assert len(sections_indices) == len(shifted)
        # start + 2 because we need to remove the header and the separator line
        ranges = [range(a + 2, b - 1) for a, b in zip(sections_indices.values(), shifted)]
        return {v:r for r, v in zip(ranges, sections_indices.keys())}

    @property
    def all_visible_sections(self) -> List[str]:
        """
        Returns a list of all visible sections
        """
        return list(self.all_visible_sections_ranges)

    @property
    def custom_sections(self) -> Dict[str, List[str]]:
        """
        Returns a dict of custom sections and their content (see class Examples)
        """
        sections = {}
        for section, range_ in self.all_visible_sections_ranges.items():
            if section not in self.standard_section_names:
                lines = self.docstring_lines[range_.start:range_.stop + 1]
                sections[section] = lines
        return sections

    @property
    def standard_sections(self) -> Dict[str, list]:
        """
        Returns a dictionary similar to NumpyDocString(obj).sections
        """
        return {k:self.doc[k] for k in self.standard_section_names}

    @property
    def all_sections(self) -> Dict[str, list]:
        """
        Returns a dictionary with standard sections from numpydoc AND
        custom sections we found (see class Examples)
        """
        # we need to iterate through all sections to preserve the docstring order
        # instead of just adding properties custom_sections and sections
        standard_sections = self.standard_sections
        custom_sections = self.custom_sections
        all_visible_sections = self.all_visible_sections

        # case when custom sections cannot be positionned relative to standard
        # ones because there are no visible standard sections
        # in this case we'll put custom sections after "Extended Summary"
        only_custom = all(v in custom_sections for v in all_visible_sections)
        if only_custom:
            default_keys = list(standard_sections.keys())
            default_values = list(standard_sections.values())
            ix_custom = default_keys.index('Extended Summary') + 1

            keys = default_keys[0:ix_custom] + list(custom_sections.keys()) + default_keys[ix_custom:]
            values = default_values[0:ix_custom] + list(custom_sections.values()) + default_values[ix_custom:]
            return dict(zip(keys, values))

        # case when custom sections can be positionned relative to standard ones
        # local helper
        def get_consecutive_custom_section(section_name:str) -> Union[None, List[str]]:
            if section_name in all_visible_sections:
                ix = all_visible_sections.index(section_name) + 1
                subset = all_visible_sections[ix:]
                return next((k for k in subset if k in custom_sections), None)
            else:
                return None

        # get all sections in order
        all_sections = {}
        for k, v in standard_sections.items():
            all_sections[k] = v
            # check if a custom section follows the standard section
            consecutive_custom_section = get_consecutive_custom_section(k)
            if consecutive_custom_section is not None:
                v_custom = custom_sections[consecutive_custom_section]
                all_sections[consecutive_custom_section] = v_custom
        return all_sections

    def __getitem__(self, key:str):
        # that's not particulary efficient since the property
        # will be computed each time but it's not a big deal
        return self.all_sections[key]



# -*- coding: utf-8 -*-
import setuptools
import os

here = os.path.abspath(os.path.dirname(__file__))

description = 'Automatic documentation tool using Markdown and Python docstrings written with the numpydoc style'

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = description


with open(os.path.join(here, "requirements.txt"),"r") as f:
    requirements = [line.strip() for line in f.readlines()]

setuptools.setup(
    name="npdoc_to_md",
    version="2.0",
    license='The Unlicense',
    author="Thibault Bétrémieux",
    author_email="thibault.betremieux@gmail.com",
    url='https://github.com/ThibTrip/npdoc_to_md',
    download_url='https://github.com/ThibTrip/npdoc_to_md/archive/v2.0.tar.gz',
    keywords=['numpydoc','documentation', 'docstrings', 'python', 'markdown'],
    entry_points={'console_scripts': ['npdoc-to-md=npdoc_to_md.core:start_cli',
                                      # aliases (note that python-fire, the library
                                      # we use for the CLI also allows mixing "-"
                                      # and "_" in the same argument)
                                      'npdoc_to_md=npdoc_to_md.core:start_cli',
                                      'npdoc-to_md=npdoc_to_md.core:start_cli',
                                      'npdoc_to-md=npdoc_to_md.core:start_cli']},
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    classifiers=["Development Status :: 5 - Production/Stable",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "License :: Public Domain",
                 "Intended Audience :: Developers",
                 "Operating System :: OS Independent"])



"""
Expectations of how the docstrings of test objects in this library
should be rendered.

I put this in a separate file so that our tests aren't cluttered.

IMPORTANT: all the rendered docstrings must be Python raw strings!

See module `test_core.py`
"""

# # Common rendering options that were used to obtain the results below
#
# The example with public methods of a class will additionally use `members=['$public']` (see `test_core` module)

RENDERING_OPTIONS = dict(remove_doctest_blanklines=True,
                         remove_doctest_skip=True,
                         examples_md_lang='raw')

# # Rendered result of `documented_func_example`

documented_func_example_md = r"""
## <span style="color:purple">npdoc\_to\_md.testing.documented\_func\_example</span>_(a: int, b: int, \*, useless\_param=None, \_bad\_name\_param1=None, \_\_bad\_name\_param2=None, crazy\_typing\_param: Union[Dict[str, int], List[str]] = None, \*\*kwargs) -> int_

Example function that sums to integers. This function is used for pytest.
It does not really make sense, this is just for testing docstring rendering.

Some extended summary.

Multiline test.

### Parameters

* **a** : **_int_**

  First number.

  Multiline test.

* **b** : **_int_**

### Returns

* **c** : **_int_**

  Result of a + b.

  Multiline test.

### Raises

* **_TypeError_**

  If a or b is not a number.

  Multiline test.

### Warns

* **_UserWarning_**

  Never occurs (this is just an example)

  Multiline test.

### Other Parameters

* **useless\_param**

  Some useless parameter.

  Multiline test.

* **\_bad\_name\_param1**

* **\_\_bad\_name\_param2**

* **crazy\_typing\_param**

  Parameter with complex typing

* **optional\_type\_param**

  Parameter with optional typing

### See Also

* **empty\_function, EmptyClass, empty\_doc\_sections\_function, EmptyDocSectionsClass**

  Other useless functions.
  Numpydoc deletes line breaks for section "See Also" so no multiline test here...
  The next function I put in this section will have no description.

* **now\_utc**

### Notes

Some notes.

Multiline test.

### Warnings

Cautionnary notes for the user.
More warning notes.

Multiline test.

### References

[0] Me

Multiline test.

### Examples

We are going to try to break our library by:
1. Putting text before examples or not
2. Separating examples by line breaks (also with irrelevant leading spaces) or not

We'll then check if the library still manages to separate examples properly

* example 1
```python
documented_func_example(1, 1)
```
```python
2
```

* example 2
```python
documented_func_example(2, 2)
```
```python
4
```

---

#### Examples unrelated to the function

* test rendered markdown
```python
import pandas as pd
df = pd.DataFrame({'A':[1, 2]})
print(df.to_markdown())
```
|    |   A |
|---:|----:|
|  0 |   1 |
|  1 |   2 |

* test raw code block
```python
print('Hello world!')
```
```
Hello world!
```

* test non rendered markdown
```python
print('# some markdown string\n* test')
```
```markdown
# some markdown string
* test
```

```python
print('test')
```
```
test
```
```python
print('foo\n')
```
```
foo

```
""".strip()

# # Rendered result of `documented_generator_func_example`

documented_generator_func_example_md = r"""
## <span style="color:purple">npdoc\_to\_md.testing.documented\_generator\_func\_example</span>_(nb\_vals=10)_

Generator example (for testing the parsing of Yields and Receives sections).
Simply yields a range from 0 to `nb_vals`.

### Yields

* **i** : **_int_**

  Incrementing integers.

  Multiline test.

### Receives

* **nb\_vals** : **_int, default 10_**

  Number of values to yield.

  Multiline test.

### Examples

```python
for i in documented_generator_func_example(2):
   print(i)
```
```
0
1
```""".strip()

# # Rendered result of `DocumentedClassExample` with public methods

DocumentedClassExampleMd = r"""
## <span style="color:purple">npdoc\_to\_md.testing.DocumentedClassExample</span>_(first\_name: str, last\_name: str) -> None_

Example dataclass representing a person. This class is used for pytest.
It does not really make sense, this is just for testing docstring rendering.

### Attributes

* **full\_name** : **_str_**

  first name + last name (both names in title case and separated by a space).

  Multiline test.

### Examples

```python
ex = DocumentedClassExample(first_name='John', last_name='Doe')
ex.full_name
```
```
'John Doe'
```

## <span style="color:purple">npdoc\_to\_md.testing.DocumentedClassExample.example\_method</span>_() -> None_

Prints the first name defined in the instance.
This is a test for rendering docstrings of methods.

### Examples

```python
d = DocumentedClassExample('John', 'Doe')
d.example_method()
```
```
John
```

## <span style="color:purple">npdoc\_to\_md.testing.DocumentedClassExample.example\_static\_method</span>_() -> List[str]_

Prints the names of the dataclass fields.
This is a test for rendering docstrings of staticmethods.

### Examples

```python
DocumentedClassExample.example_static_method()
```
```
['first_name', 'last_name']
```

```python
d = DocumentedClassExample('John', 'Doe')
d.example_static_method()
```
```
['first_name', 'last_name']
```

## <span style="color:purple">npdoc\_to\_md.testing.DocumentedClassExample.full\_name</span>

Returns the full name using the names part defined in the instance.
This is a test for rendering docstrings of properties.

### Returns

* **_str_**

  Full name

### Examples

```python
ex = DocumentedClassExample(first_name='john', last_name='doe')
ex.full_name
```
```
'John Doe'
```

## <span style="color:purple">npdoc\_to\_md.testing.DocumentedClassExample.test\_class\_method</span>_() -> 'DocumentedClassExample'_

Returns a "John Doe" instance of the class.
This is a test for rendering docstrings of class methods.

### Examples

```python
DocumentedClassExample.test_class_method()
```
```
DocumentedClassExample(first_name='John', last_name='Doe')
```
""".strip()

# # Rendered result of `npdoc_to_md.testing` module

testing_module_md = r"""
## <span style="color:purple">npdoc\_to\_md.testing</span>

Dummy objects (classes, functions, ...) for demos of the library and tests

### Notes

The docstring of this module will be used for docstring rendering tests in pytest.
Which is why I am also going to include some examples here and a custom section
"Functions".

### Functions

* create_random_table
  Returns a random table (pandas DataFrame)
* now_utc
  Returns the current date and time in UTC
* ...

### Examples

```python
from npdoc_to_md.testing import DocumentedClassExample

ex = DocumentedClassExample(first_name='John', last_name='Doe')
ex.full_name
```
```
'John Doe'
```""".strip()

# # Rendered results of objects with empty docstrings or empty sections

# +
empty_function_md = r'## <span style="color:purple">npdoc\_to\_md.testing.empty\_function</span>_()_'
EmptyClassMd = r'## <span style="color:purple">npdoc\_to\_md.testing.EmptyClass</span>_()_'
EmptyClassDictMd = r'## <span style="color:purple">npdoc\_to\_md.testing.EmptyClass.\_\_dict\_\_</span>'  # render of __dict__ method's docstring
empty_doc_sections_function_md = r"""
## <span style="color:purple">npdoc\_to\_md.testing.empty\_doc\_sections\_function</span>_()_



### Parameters



### Yields



### Receives



### Raises



### Warns



### Other Parameters



### Attributes



### Methods
""".strip()

EmptyDocSectionsClass_md = r"""
## <span style="color:purple">npdoc\_to\_md.testing.EmptyDocSectionsClass</span>_()_



### Parameters



### Yields



### Receives



### Raises



### Warns



### Other Parameters



### Attributes



### Methods
""".strip()

none_md = r'## <span style="color:purple">None</span>'  # rendered version of None object
builtins_none_md = r'## <span style="color:purple">builtins.None</span>'  # rendered version of None object imported from builtins



"""
Helpers for the various modules of the library
"""
import os
import re
from pathlib import Path
from typing import Iterable, List


# # Regex patterns

class Patterns:
    """
    All regex patterns used in the library

    Attributes
    ----------
    console_py
        Matches inputs in docstring examples
    doctest_skip
        Matches "# doctest: +SKIP" marker of doctest in docstring examples
    blankline
        Matches "<BLANKLINE>" marker of doctest in docstring examples
    self_or_cls
        Matches special arguments "self" or "cls" that we remove from signatures.
        Also matches any subsequent character we'll have to remove (following space(s)
        and comma(s))
    template_files
        Matches file extensions ".np" and ".md" case sensitive
    template_files_insensitive
        Same as attribute `template_files` but case insensitive
    """
    console_py = re.compile(r'^(\>\>\> ?|\.\.\. ?)')
    doctest_skip = re.compile(r' *\# *doctest: *\+SKIP *$')
    blankline = re.compile(r'^ *\<BLANKLINE\> *$')
    self_or_cls = re.compile(r'(?<=\()(self|cls) *\,* *')  # positive lookbehind of "(" before self|cls
    template_files = re.compile(r'\.(np)?md$')
    template_files_insensitive = re.compile(template_files.pattern, flags=re.IGNORECASE)


# # Generic helpers

# +
def previous_current_next(iterable:Iterable) -> Iterable[tuple]:
    """
    Iterator that yields (previous, current, next) tuple per element.

    Returns None if the value does not make sense (i.e. previous before
    first and next after last).

    Stolen from: https://gist.github.com/mortenpi/9604377?permalink_comment_id=3180599#gistcomment-3180599

    Examples
    --------
    >>> for t in previous_current_next((1, 2, 3)):
    ...     print(t)
    (None, 1, 2)
    (1, 2, 3)
    (2, 3, None)
    """
    iterable = iter(iterable)
    prv = None
    cur = iterable.__next__()
    try:
        while True:
            nxt = iterable.__next__()
            yield (prv, cur, nxt)
            prv = cur
            cur = nxt
    except StopIteration:
        yield (prv, cur, None)


def unique(v:Iterable) -> list:
    """
    Produces a unique list from an iterable.
    This is similar to doing list(set(some_list)) but it preserves the order.

    Examples
    --------
    >>> unique(['foo', 'foo'])
    ['foo']
    """
    u = []
    for x in v:
        if x not in u:
            u.append(x)
    return u


# -

# # Helper for file operations

class FileOperations:
    """
    Collection of methods for operations relative to files and folders
    e.g. `FileOperations.list_files`.
    """

    def _list_files_recursive(folder:str, pattern:re.Pattern) -> List[str]:
        """
        Lists files whose name match `pattern` in a folder and its subfolders
        """
        filepaths = []
        for root, dirs, files in os.walk(folder):
            for f in files:
                # search pattern inside of file name
                if pattern.search(Path(f).name):
                    fullpath = os.path.join(root, f)
                    filepaths.append(fullpath)
        return filepaths

    def _list_files_non_recursive(folder:str, pattern:re.Pattern) -> List[str]:
        """
        Lists files whose name match `pattern` in a folder but **not** in its subfolders
        """
        return [os.path.join(folder, f) for f in os.listdir(folder)
                if pattern.search(Path(f).name)]

    def list_files(folder:str, pattern:re.Pattern, recursive:bool=True) -> List[str]:
        """
        Lists files whose name match `pattern` in a `folder` and optionally in its subfolders
        (when `recursive=True`)
        """
        if recursive:
            f = FileOperations._list_files_recursive
        else:
            f = FileOperations._list_files_non_recursive
        return f(folder=folder, pattern=pattern)

    def switch_folder(source_folder:str, destination_folder:str,
                      filepath:str, create_missing_dirs:bool=True) -> str:
        """
        Switches the folder where a file is located.

        It will reproduce the folder structure of `source_folder`
        in `destination_folder`.

        Examples
        --------
        >>> FileOperations.switch_folder('D:/', 'C:/Test', filepath='D:/Subfolder/foo.md', create_missing_dirs=False) # doctest: +SKIP
        'C:/Test/Subfolder/foo.md'
        """
        relpath = os.path.relpath(filepath, source_folder)
        destination_path = os.path.join(destination_folder, relpath)
        if create_missing_dirs:
            destination_folder = Path(destination_path).parent
            destination_folder.mkdir(parents=True, exist_ok=True)
        return destination_path
